"""Event categories resources"""

from flask.views import MethodView
from flask_smorest import abort

from bemserver_core.model import EventCategory

from bemserver_api import Blueprint
from bemserver_api.database import db

from .schemas import EventCategorySchema


blp = Blueprint(
    "EventCategory",
    __name__,
    url_prefix="/event_categories",
    description="Operations on event categories",
)


@blp.route("/")
class EventCategoriesViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.response(200, EventCategorySchema(many=True))
    def get(self):
        """List event categories"""
        return EventCategory.get()

    @blp.login_required
    @blp.etag
    @blp.arguments(EventCategorySchema)
    @blp.response(201, EventCategorySchema)
    @blp.catch_integrity_error
    def post(self, new_item):
        """Add a new event"""
        item = EventCategory.new(**new_item)
        db.session.commit()
        return item


@blp.route("/<string:item_id>")
class EventCategoriesByIdViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.response(200, EventCategorySchema)
    def get(self, item_id):
        """Get en event by its ID"""
        item = EventCategory.get_by_id(item_id)
        if item is None:
            abort(404)
        return item

    @blp.login_required
    @blp.etag
    @blp.arguments(EventCategorySchema)
    @blp.response(200, EventCategorySchema)
    @blp.catch_integrity_error
    def put(self, new_item, item_id):
        """Update an existing timeseries"""
        item = EventCategory.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, EventCategorySchema)
        item.update(**new_item)
        db.session.commit()
        return item

    @blp.login_required
    @blp.etag
    @blp.response(204)
    def delete(self, item_id):
        """Delete an event"""
        item = EventCategory.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, EventCategorySchema)
        item.delete()
        db.session.commit()



"""Campaign scopes resources"""
from flask.views import MethodView
from flask_smorest import abort

from bemserver_core.model import CampaignScope

from bemserver_api import Blueprint
from bemserver_api.database import db

from .schemas import (
    CampaignScopeSchema,
    CampaignScopePutSchema,
    CampaignScopeQueryArgsSchema,
)


blp = Blueprint(
    "CampaignScope",
    __name__,
    url_prefix="/campaign_scopes",
    description="Operations on campaign scopes",
)


@blp.route("/")
class CampaignScopeViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.arguments(CampaignScopeQueryArgsSchema, location="query")
    @blp.response(200, CampaignScopeSchema(many=True))
    def get(self, args):
        """List campaign scopes"""
        return CampaignScope.get(**args)

    @blp.login_required
    @blp.etag
    @blp.arguments(CampaignScopeSchema)
    @blp.response(201, CampaignScopeSchema)
    @blp.catch_integrity_error
    def post(self, new_item):
        """Add a new campaign scope"""
        item = CampaignScope.new(**new_item)
        db.session.commit()
        return item


@blp.route("/<int:item_id>")
class CampaignScopeByIdViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.response(200, CampaignScopeSchema)
    def get(self, item_id):
        """Get campaign scope by ID"""
        item = CampaignScope.get_by_id(item_id)
        if item is None:
            abort(404)
        return item

    @blp.login_required
    @blp.etag
    @blp.arguments(CampaignScopePutSchema)
    @blp.response(200, CampaignScopeSchema)
    @blp.catch_integrity_error
    def put(self, new_item, item_id):
        """Update an existing campaign scope"""
        item = CampaignScope.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, CampaignScopeSchema)
        item.update(**new_item)
        db.session.commit()
        return item

    @blp.login_required
    @blp.etag
    @blp.response(204)
    def delete(self, item_id):
        """Delete a campaign scope"""
        item = CampaignScope.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, CampaignScopeSchema)
        item.delete()
        db.session.commit()



"""Space properties resources"""
from flask.views import MethodView
from flask_smorest import abort

from bemserver_core.model import SpaceProperty

from bemserver_api import Blueprint
from bemserver_api.database import db

from .schemas import (
    SpacePropertySchema,
    SpacePropertyQueryArgsSchema,
)


blp = Blueprint(
    "SpaceProperty",
    __name__,
    url_prefix="/space_properties",
    description="Operations on space properties",
)


@blp.route("/")
class SpacePropertyViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.arguments(SpacePropertyQueryArgsSchema, location="query")
    @blp.response(200, SpacePropertySchema(many=True))
    def get(self, args):
        """List space properties"""
        return SpaceProperty.get(**args)

    @blp.login_required
    @blp.etag
    @blp.arguments(SpacePropertySchema)
    @blp.response(201, SpacePropertySchema)
    @blp.catch_integrity_error
    def post(self, new_item):
        """Add a new space property"""
        item = SpaceProperty.new(**new_item)
        db.session.commit()
        return item


@blp.route("/<int:item_id>")
class SpacePropertyByIdViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.response(200, SpacePropertySchema)
    def get(self, item_id):
        """Get space property by ID"""
        item = SpaceProperty.get_by_id(item_id)
        if item is None:
            abort(404)
        return item

    @blp.login_required
    @blp.response(204)
    def delete(self, item_id):
        """Delete a space property"""
        item = SpaceProperty.get_by_id(item_id)
        if item is None:
            abort(404)
        item.delete()
        db.session.commit()



"""Event levels tests"""
import pytest

from tests.common import AuthHeader


EVENT_LEVELS_URL = "/event_levels/"


class TestEventLevelsApi:
    @pytest.mark.parametrize("user", ("user", "admin"))
    def test_event_levels_api_as_admin_or_user(self, app, user, users):

        if user == "user":
            creds = users["Active"]["creds"]
        else:
            creds = users["Chuck"]["creds"]

        client = app.test_client()

        with AuthHeader(creds):

            # GET level list
            ret = client.get(EVENT_LEVELS_URL)
            assert ret.status_code == 200
            assert len(ret.json)

    def test_event_levels_api_as_anonym(self, app):

        client = app.test_client()

        ret = client.get(EVENT_LEVELS_URL)
        assert ret.status_code == 401



"""Test integrity error"""
import sqlalchemy as sqla
import psycopg2.errors as ppe
import flask

import pytest

from bemserver_api import Api, Blueprint


uve = sqla.exc.IntegrityError(None, None, ppe.UniqueViolation())
fkve = sqla.exc.IntegrityError(None, None, ppe.ForeignKeyViolation())
uve = sqla.exc.IntegrityError(None, None, None)


class TestIntegrityError:
    @pytest.mark.parametrize("error", (uve, fkve, uve))
    def test_blp_integrity_error(self, error):

        app = flask.Flask("Test")
        api = Api(
            app, spec_kwargs={"title": "Test", "version": "1", "openapi_version": "3"}
        )

        blp = Blueprint("Test", __name__, url_prefix="/test")

        @blp.route("/decorator")
        @blp.response(200)
        @blp.catch_integrity_error()
        def decorator():
            raise error

        @blp.route("/decorator_factory")
        @blp.response(200)
        @blp.catch_integrity_error
        def decorator_factory():
            raise error

        api.register_blueprint(blp)
        client = app.test_client()

        resp = client.get("/test/decorator")
        assert resp.status_code == 409
        resp = client.get("/test/decorator_factory")
        assert resp.status_code == 409



"""ST_CleanupByTimeseries resources"""
from flask.views import MethodView
from flask_smorest import abort

from bemserver_core.scheduled_tasks import ST_CleanupByTimeseries

from bemserver_api import Blueprint

from .schemas import (
    ST_CleanupByTimeseriesSchema,
    ST_CleanupByTimeseriesQueryArgsSchema,
    ST_CleanupByTimeseriesFullSchema,
    ST_CleanupByTimeseriesFullQueryArgsSchema,
)


blp = Blueprint(
    "ST_CleanupByTimeseries",
    __name__,
    url_prefix="/st_cleanups_by_timeseries",
    description="Operations on cleanup scheduled task x timeseries associations",
)


@blp.route("/")
class ST_CleanupByTimeseriesViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.arguments(ST_CleanupByTimeseriesQueryArgsSchema, location="query")
    @blp.response(200, ST_CleanupByTimeseriesSchema(many=True))
    def get(self, args):
        """List cleanup scheduled task x timeseries associations"""
        return ST_CleanupByTimeseries.get(**args)


@blp.route("/<int:item_id>")
class ST_CleanupByTimeseriesByIdViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.response(200, ST_CleanupByTimeseriesSchema)
    def get(self, item_id):
        """Get cleanup scheduled task x timeseries association by ID"""
        item = ST_CleanupByTimeseries.get_by_id(item_id)
        if item is None:
            abort(404)
        return item


@blp.route("/full")
@blp.login_required
@blp.etag
@blp.arguments(ST_CleanupByTimeseriesFullQueryArgsSchema, location="query")
@blp.response(200, ST_CleanupByTimeseriesFullSchema(many=True))
def get_full(args):
    """List cleanup service last timestamp for all timeseries"""
    return ST_CleanupByTimeseries.get_all(**args)



"""Timeseries by zones routes tests"""
import pytest

from tests.common import AuthHeader


DUMMY_ID = "69"

TIMESERIES_BY_ZONES_URL = "/timeseries_by_zones/"
ZONES_URL = "/zones/"


class TestTimeseriesByZoneApi:
    def test_timeseries_by_zones_api(self, app, users, zones, timeseries):

        creds = users["Chuck"]["creds"]
        zone_1_id = zones[0]
        zone_2_id = zones[1]
        ts_1_id = timeseries[0]
        ts_2_id = timeseries[1]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(TIMESERIES_BY_ZONES_URL)
            assert ret.status_code == 200
            assert ret.json == []

            # POST
            tbz_1 = {
                "zone_id": zone_1_id,
                "timeseries_id": ts_1_id,
            }
            ret = client.post(TIMESERIES_BY_ZONES_URL, json=tbz_1)
            assert ret.status_code == 201
            ret_val = ret.json
            tbz_1_id = ret_val.pop("id")
            tbz_1_etag = ret.headers["ETag"]
            assert ret_val == tbz_1

            # POST violating unique constraint
            ret = client.post(TIMESERIES_BY_ZONES_URL, json=tbz_1)
            assert ret.status_code == 409

            # GET list
            ret = client.get(TIMESERIES_BY_ZONES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == ts_1_id

            # GET by id
            ret = client.get(f"{TIMESERIES_BY_ZONES_URL}{tbz_1_id}")
            assert ret.status_code == 200
            assert ret.headers["ETag"] == tbz_1_etag
            ret_val = ret.json
            ret_val.pop("id")
            assert ret_val == tbz_1

            # PUT
            tbz_1["timeseries_id"] = ts_2_id
            ret = client.put(
                f"{TIMESERIES_BY_ZONES_URL}{tbz_1_id}",
                json=tbz_1,
                headers={"If-Match": tbz_1_etag},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            ret_val.pop("id")
            tbz_1_etag = ret.headers["ETag"]
            assert ret_val == tbz_1

            # PUT wrong ID -> 404
            ret = client.put(
                f"{TIMESERIES_BY_ZONES_URL}{DUMMY_ID}",
                json=tbz_1,
                headers={"If-Match": tbz_1_etag},
            )
            assert ret.status_code == 404

            # POST sep 2
            tbz_2 = {
                "zone_id": zone_2_id,
                "timeseries_id": ts_2_id,
            }
            ret = client.post(TIMESERIES_BY_ZONES_URL, json=tbz_2)
            ret_val = ret.json
            tbz_2_id = ret_val.pop("id")
            tbz_2_etag = ret.headers["ETag"]

            # PUT violating unique constraint
            tbz_1["zone_id"] = tbz_2["zone_id"]
            tbz_1["timeseries_id"] = tbz_2["timeseries_id"]
            ret = client.put(
                f"{TIMESERIES_BY_ZONES_URL}{tbz_1_id}",
                json=tbz_1,
                headers={"If-Match": tbz_1_etag},
            )
            assert ret.status_code == 409

            # GET list
            ret = client.get(TIMESERIES_BY_ZONES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2

            # GET list with filters
            ret = client.get(
                TIMESERIES_BY_ZONES_URL,
                query_string={"zone_id": zone_1_id},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == tbz_1_id

            # DELETE wrong ID -> 404
            ret = client.delete(
                f"{TIMESERIES_BY_ZONES_URL}{DUMMY_ID}",
                headers={"If-Match": tbz_1_etag},
            )
            assert ret.status_code == 404

            # DELETE zone cascade
            ret = client.get(f"{ZONES_URL}{zone_1_id}")
            zone_1_etag = ret.headers["ETag"]
            ret = client.delete(
                f"{ZONES_URL}{zone_1_id}", headers={"If-Match": zone_1_etag}
            )
            assert ret.status_code == 204

            # DELETE
            ret = client.delete(
                f"{TIMESERIES_BY_ZONES_URL}{tbz_1_id}",
                headers={"If-Match": tbz_1_etag},
            )
            assert ret.status_code == 404
            ret = client.delete(
                f"{TIMESERIES_BY_ZONES_URL}{tbz_2_id}",
                headers={"If-Match": tbz_2_etag},
            )
            assert ret.status_code == 204

            # GET list
            ret = client.get(TIMESERIES_BY_ZONES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # GET by id -> 404
            ret = client.get(f"{TIMESERIES_BY_ZONES_URL}{tbz_1_id}")
            assert ret.status_code == 404

    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaigns")
    @pytest.mark.usefixtures("user_groups_by_campaign_scopes")
    @pytest.mark.parametrize("timeseries", (4,), indirect=True)
    def test_timeseries_by_zones_as_user_api(
        self, app, users, zones, timeseries, timeseries_by_zones
    ):

        user_creds = users["Active"]["creds"]
        zone_1_id = zones[0]
        ts_1_id = timeseries[0]
        ts_3_id = timeseries[2]
        tbz_1_id = timeseries_by_zones[0]

        client = app.test_client()

        with AuthHeader(user_creds):

            # GET list
            ret = client.get(TIMESERIES_BY_ZONES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            tbz_1 = ret_val[0]
            assert tbz_1.pop("id") == tbz_1_id

            # POST
            tbz_3 = {
                "zone_id": zone_1_id,
                "timeseries_id": ts_1_id,
            }
            ret = client.post(TIMESERIES_BY_ZONES_URL, json=tbz_3)
            assert ret.status_code == 403

            # GET by id
            ret = client.get(f"{TIMESERIES_BY_ZONES_URL}{tbz_1_id}")
            assert ret.status_code == 200
            tbz_1_etag = ret.headers["ETag"]

            # PUT
            tbz_1["timeseries_id"] = ts_3_id
            ret = client.put(
                f"{TIMESERIES_BY_ZONES_URL}{tbz_1_id}",
                json=tbz_1,
                headers={"If-Match": tbz_1_etag},
            )
            assert ret.status_code == 403

            # DELETE
            ret = client.delete(
                f"{TIMESERIES_BY_ZONES_URL}{tbz_1_id}",
                headers={"If-Match": tbz_1_etag},
            )
            assert ret.status_code == 403

    def test_timeseries_by_zones_as_anonym_api(
        self, app, zones, timeseries, timeseries_by_zones
    ):
        zone_1_id = zones[0]
        ts_1_id = timeseries[0]
        ts_2_id = timeseries[1]
        tbz_1_id = timeseries_by_zones[0]

        client = app.test_client()

        # GET list
        ret = client.get(TIMESERIES_BY_ZONES_URL)
        assert ret.status_code == 401

        # POST
        tbz_3 = {
            "zone_id": zone_1_id,
            "timeseries_id": ts_2_id,
        }
        ret = client.post(TIMESERIES_BY_ZONES_URL, json=tbz_3)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(f"{TIMESERIES_BY_ZONES_URL}{tbz_1_id}")
        assert ret.status_code == 401

        # PUT
        tbz_1 = {
            "zone_id": zone_1_id,
            "timeseries_id": ts_1_id,
        }
        ret = client.put(
            f"{TIMESERIES_BY_ZONES_URL}{tbz_1_id}",
            json=tbz_1,
            headers={"If-Match": "Dummy-ETag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401

        # DELETE
        ret = client.delete(
            f"{TIMESERIES_BY_ZONES_URL}{tbz_1_id}",
            headers={"If-Match": "Dummy-Etag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401



"""Structural element properties API schemas"""

import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.model import StructuralElementProperty
from bemserver_core.common import PropertyType

from bemserver_api import AutoSchema, Schema, SortField


class StructuralElementPropertySchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = StructuralElementProperty

    id = msa.auto_field(dump_only=True)
    name = msa.auto_field(validate=ma.validate.Length(1, 80))
    value_type = ma.fields.Enum(PropertyType, metadata={"default": "string"})


class StructuralElementPropertyPutSchema(StructuralElementPropertySchema):
    class Meta(StructuralElementPropertySchema.Meta):
        exclude = ("value_type",)


class StructuralElementPropertyQueryArgsSchema(Schema):
    sort = SortField(("name",))
    name = ma.fields.Str()
    value_type = ma.fields.Enum(PropertyType)
    unit_symbol = ma.fields.Str()



"""Test utils"""
import sqlalchemy as sqla
import pandas as pd

from bemserver_core.model import (
    TimeseriesData,
)
from bemserver_core.database import db
from bemserver_core.authorization import OpenBar


def create_timeseries_data(timeseries, data_state, timestamps, values):
    """Create timeseries data

    :param Timeseries timeseries: Timeseries
    :param TimeseriesDataState data_state: Timeseries data state
    :param list timestamps: List of timestamps
    :param list values: List of values

    timestamps and values must be of same length
    """

    with OpenBar():

        tsbds = timeseries.get_timeseries_by_data_state(data_state)

        in_df = pd.DataFrame(
            {tsbds.id: values}, index=pd.DatetimeIndex(timestamps, name="timestamp")
        )
        in_df = in_df.melt(
            value_vars=in_df.columns,
            var_name="ts_by_data_state_id",
            ignore_index=False,
        )
        data_rows = [
            row
            for row in in_df.reset_index().to_dict(orient="records")
            if pd.notna(row["value"])
        ]
        # Empty data_rows would result in cryptic error later
        assert data_rows
        query = (
            sqla.dialects.postgresql.insert(TimeseriesData)
            .values(data_rows)
            .on_conflict_do_nothing()
        )
        db.session.execute(query)
        db.session.commit()



"""Timeseries data states resources"""

from flask.views import MethodView
from flask_smorest import abort

from bemserver_core.model import TimeseriesDataState

from bemserver_api import Blueprint
from bemserver_api.database import db

from .schemas import TimeseriesDataStateSchema


blp = Blueprint(
    "TimeseriesDataState",
    __name__,
    url_prefix="/timeseries_data_states",
    description="Operations on timeseries data states",
)


@blp.route("/")
class TimeseriesDataStatesViews(MethodView):
    @blp.login_required
    @blp.response(200, TimeseriesDataStateSchema(many=True))
    def get(self):
        """List timeseries data states"""
        return TimeseriesDataState.get()

    @blp.login_required
    @blp.etag
    @blp.arguments(TimeseriesDataStateSchema)
    @blp.response(201, TimeseriesDataStateSchema)
    @blp.catch_integrity_error
    def post(self, new_item):
        """Add a new timeseries data state"""
        item = TimeseriesDataState.new(**new_item)
        db.session.commit()
        return item


@blp.route("/<int:item_id>")
class TimeseriesDataStateByIdViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.response(200, TimeseriesDataStateSchema)
    def get(self, item_id):
        """Get timeseries data state by ID"""
        item = TimeseriesDataState.get_by_id(item_id)
        if item is None:
            abort(404)
        return item

    @blp.login_required
    @blp.etag
    @blp.arguments(TimeseriesDataStateSchema)
    @blp.response(200, TimeseriesDataStateSchema)
    @blp.catch_integrity_error
    def put(self, new_item, item_id):
        """Update an existing timeseries data state"""
        item = TimeseriesDataState.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, TimeseriesDataStateSchema)
        item.update(**new_item)
        db.session.commit()
        return item

    @blp.login_required
    @blp.etag
    @blp.response(204)
    def delete(self, item_id):
        """Delete a timeseries data state"""
        item = TimeseriesDataState.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, TimeseriesDataStateSchema)
        item.delete()
        db.session.commit()



"""User groups by campaign scopes resources"""
from flask.views import MethodView
from flask_smorest import abort

from bemserver_core.model import UserGroupByCampaignScope

from bemserver_api import Blueprint
from bemserver_api.database import db

from .schemas import (
    UserGroupByCampaignScopeSchema,
    UserGroupByCampaignScopeQueryArgsSchema,
)


blp = Blueprint(
    "UserGroupByCampaignScope",
    __name__,
    url_prefix="/user_groups_by_campaign_scopes",
    description="Operations on user group x campaign scope associations",
)


@blp.route("/")
class UserGroupByCampaignScopeViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.arguments(UserGroupByCampaignScopeQueryArgsSchema, location="query")
    @blp.response(200, UserGroupByCampaignScopeSchema(many=True))
    def get(self, args):
        """List user group x campaign scope associations"""
        return UserGroupByCampaignScope.get(**args)

    @blp.login_required
    @blp.etag
    @blp.arguments(UserGroupByCampaignScopeSchema)
    @blp.response(201, UserGroupByCampaignScopeSchema)
    @blp.catch_integrity_error
    def post(self, new_item):
        """Add a new user group x campaign scope association"""
        item = UserGroupByCampaignScope.new(**new_item)
        db.session.commit()
        return item


@blp.route("/<int:item_id>")
class UserGroupByCampaignScopeByIdViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.response(200, UserGroupByCampaignScopeSchema)
    def get(self, item_id):
        """Get user group x campaign scope association by ID"""
        item = UserGroupByCampaignScope.get_by_id(item_id)
        if item is None:
            abort(404)
        return item

    @blp.login_required
    @blp.response(204)
    def delete(self, item_id):
        """Delete a user group x campaign scope association"""
        item = UserGroupByCampaignScope.get_by_id(item_id)
        if item is None:
            abort(404)
        item.delete()
        db.session.commit()



"""Sites routes tests"""
import pytest

from tests.common import AuthHeader


DUMMY_ID = "69"

SITES_URL = "/sites/"


class TestSitesApi:
    def test_sites_api(self, app, users, campaigns):

        creds = users["Chuck"]["creds"]
        campaign_1_id = campaigns[0]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(SITES_URL)
            assert ret.status_code == 200
            assert ret.json == []

            # POST
            site_1 = {
                "name": "Site 1",
                "campaign_id": campaign_1_id,
            }
            ret = client.post(SITES_URL, json=site_1)
            assert ret.status_code == 201
            ret_val = ret.json
            site_1_id = ret_val.pop("id")
            site_1_etag = ret.headers["ETag"]
            assert ret_val == site_1

            # POST violating unique constraint
            ret = client.post(SITES_URL, json=site_1)
            assert ret.status_code == 409

            # GET list
            ret = client.get(SITES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == site_1_id

            # GET by id
            ret = client.get(f"{SITES_URL}{site_1_id}")
            assert ret.status_code == 200
            assert ret.headers["ETag"] == site_1_etag
            ret_val = ret.json
            ret_val.pop("id")
            assert ret_val == site_1

            # PUT
            site_1["description"] = "Fantastic site"
            site_1_put = site_1.copy()
            del site_1_put["campaign_id"]
            ret = client.put(
                f"{SITES_URL}{site_1_id}",
                json=site_1_put,
                headers={"If-Match": site_1_etag},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            ret_val.pop("id")
            site_1_etag = ret.headers["ETag"]
            assert ret_val == site_1

            # PUT wrong ID -> 404
            ret = client.put(
                f"{SITES_URL}{DUMMY_ID}",
                json=site_1_put,
                headers={"If-Match": site_1_etag},
            )
            assert ret.status_code == 404

            # POST site 2
            site_2 = {
                "name": "Site 2",
                "campaign_id": campaign_1_id,
            }
            ret = client.post(SITES_URL, json=site_2)
            ret_val = ret.json
            site_2_id = ret_val.pop("id")
            site_2_etag = ret.headers["ETag"]

            # PUT violating unique constraint
            site_1_put["name"] = site_2["name"]
            ret = client.put(
                f"{SITES_URL}{site_1_id}",
                json=site_1_put,
                headers={"If-Match": site_1_etag},
            )
            assert ret.status_code == 409

            # GET list
            ret = client.get(SITES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2

            # GET list with filters
            ret = client.get(SITES_URL, query_string={"name": "Site 1"})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == site_1_id

            # DELETE wrong ID -> 404
            ret = client.delete(
                f"{SITES_URL}{DUMMY_ID}", headers={"If-Match": site_1_etag}
            )
            assert ret.status_code == 404

            # DELETE
            ret = client.delete(
                f"{SITES_URL}{site_1_id}", headers={"If-Match": site_1_etag}
            )
            assert ret.status_code == 204
            ret = client.delete(
                f"{SITES_URL}{site_2_id}", headers={"If-Match": site_2_etag}
            )
            assert ret.status_code == 204

            # GET list
            ret = client.get(SITES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # GET by id -> 404
            ret = client.get(f"{SITES_URL}{site_1_id}")
            assert ret.status_code == 404

    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaigns")
    def test_sites_as_user_api(self, app, users, campaigns, sites):

        user_creds = users["Active"]["creds"]
        campaign_2_id = campaigns[1]
        site_1_id = sites[0]
        site_2_id = sites[1]

        client = app.test_client()

        with AuthHeader(user_creds):

            # GET list
            ret = client.get(SITES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            site_1 = ret_val[0]
            assert site_1.pop("id") == site_1_id

            # GET list with filters
            ret = client.get(SITES_URL, query_string={"name": "Site 1"})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == site_1_id
            ret = client.get(SITES_URL, query_string={"name": "Site 2"})
            assert ret.status_code == 200
            assert not ret.json

            # POST
            site_3 = {
                "name": "Site 3",
                "campaign_id": campaign_2_id,
            }
            ret = client.post(SITES_URL, json=site_3)
            assert ret.status_code == 403

            # GET by id
            ret = client.get(f"{SITES_URL}{site_1_id}")
            assert ret.status_code == 200
            site_1_etag = ret.headers["ETag"]

            ret = client.get(f"{SITES_URL}{site_2_id}")
            assert ret.status_code == 403

            # PUT
            site_1["description"] = "Fantastic site"
            site_1_put = site_1.copy()
            del site_1_put["campaign_id"]
            ret = client.put(
                f"{SITES_URL}{site_1_id}",
                json=site_1_put,
                headers={"If-Match": site_1_etag},
            )
            assert ret.status_code == 403

            # DELETE
            ret = client.delete(
                f"{SITES_URL}{site_1_id}", headers={"If-Match": site_1_etag}
            )
            assert ret.status_code == 403

    def test_sites_as_anonym_api(self, app, sites, campaigns):

        site_1_id = sites[0]
        campaign_1_id = campaigns[0]

        client = app.test_client()

        # GET list
        ret = client.get(SITES_URL)
        assert ret.status_code == 401

        # POST
        site_3 = {
            "name": "Site 3",
            "campaign_id": campaign_1_id,
        }
        ret = client.post(SITES_URL, json=site_3)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(f"{SITES_URL}{site_1_id}")
        assert ret.status_code == 401

        # PUT
        site_1 = {
            "name": "Super Site 1",
        }
        ret = client.put(
            f"{SITES_URL}{site_1_id}",
            json=site_1,
            headers={"If-Match": "Dummy-ETag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401

        # DELETE
        ret = client.delete(
            f"{SITES_URL}{site_1_id}", headers={"If-Match": "Dummy-Etag"}
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401



"""Energy consumption timeseries by buildings routes tests"""
import pytest

from tests.common import AuthHeader


DUMMY_ID = "69"

ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL = (
    "/energy_consumption_timeseries_by_buildings/"
)
BUILDINGS_URL = "/buildings/"


class TestEnergyConsumptionTimeseriesByBuildingApi:
    def test_energy_consumption_timeseries_by_buildings_api(
        self, app, users, buildings, timeseries, energy_sources, energy_end_uses
    ):

        creds = users["Chuck"]["creds"]
        building_1_id = buildings[0]
        building_2_id = buildings[1]
        ts_1_id = timeseries[0]
        ts_2_id = timeseries[1]
        energy_source_1_id = energy_sources[1]
        energy_end_use_1_id = energy_end_uses[1]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL)
            assert ret.status_code == 200
            assert ret.json == []

            # POST
            ectbs_1 = {
                "building_id": building_1_id,
                "timeseries_id": ts_1_id,
                "source_id": energy_source_1_id,
                "end_use_id": energy_end_use_1_id,
                "wh_conversion_factor": 1,
            }
            ret = client.post(
                ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL, json=ectbs_1
            )
            assert ret.status_code == 201
            ret_val = ret.json
            ectbs_1_id = ret_val.pop("id")
            ectbs_1_etag = ret.headers["ETag"]
            assert ret_val == ectbs_1

            # POST violating unique constraint
            ret = client.post(
                ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL, json=ectbs_1
            )
            assert ret.status_code == 409

            # GET list
            ret = client.get(ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == ts_1_id

            # GET by id
            ret = client.get(
                f"{ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL}{ectbs_1_id}"
            )
            assert ret.status_code == 200
            assert ret.headers["ETag"] == ectbs_1_etag
            ret_val = ret.json
            ret_val.pop("id")
            assert ret_val == ectbs_1

            # PUT
            ectbs_1["timeseries_id"] = ts_2_id
            ret = client.put(
                f"{ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL}{ectbs_1_id}",
                json=ectbs_1,
                headers={"If-Match": ectbs_1_etag},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            ret_val.pop("id")
            ectbs_1_etag = ret.headers["ETag"]
            assert ret_val == ectbs_1

            # PUT wrong ID -> 404
            ret = client.put(
                f"{ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL}{DUMMY_ID}",
                json=ectbs_1,
                headers={"If-Match": ectbs_1_etag},
            )
            assert ret.status_code == 404

            # POST sep 2
            ectbs_2 = {
                "building_id": building_2_id,
                "timeseries_id": ts_2_id,
                "source_id": energy_source_1_id,
                "end_use_id": energy_end_use_1_id,
                "wh_conversion_factor": 1,
            }
            ret = client.post(
                ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL, json=ectbs_2
            )
            ret_val = ret.json
            ectbs_2_id = ret_val.pop("id")
            ectbs_2_etag = ret.headers["ETag"]

            # PUT violating unique constraint
            ectbs_1["building_id"] = ectbs_2["building_id"]
            ectbs_1["timeseries_id"] = ectbs_2["timeseries_id"]
            ret = client.put(
                f"{ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL}{ectbs_1_id}",
                json=ectbs_1,
                headers={"If-Match": ectbs_1_etag},
            )
            assert ret.status_code == 409

            # GET list
            ret = client.get(ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2

            # GET list with filters
            ret = client.get(
                ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL,
                query_string={"building_id": building_1_id},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == ectbs_1_id

            # DELETE wrong ID -> 404
            ret = client.delete(
                f"{ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL}{DUMMY_ID}",
                headers={"If-Match": ectbs_1_etag},
            )
            assert ret.status_code == 404

            # DELETE building cascade
            ret = client.get(f"{BUILDINGS_URL}{building_1_id}")
            building_1_etag = ret.headers["ETag"]
            ret = client.delete(
                f"{BUILDINGS_URL}{building_1_id}", headers={"If-Match": building_1_etag}
            )
            assert ret.status_code == 204

            # DELETE
            ret = client.delete(
                f"{ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL}{ectbs_1_id}",
                headers={"If-Match": ectbs_1_etag},
            )
            assert ret.status_code == 404
            ret = client.delete(
                f"{ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL}{ectbs_2_id}",
                headers={"If-Match": ectbs_2_etag},
            )
            assert ret.status_code == 204

            # GET list
            ret = client.get(ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # GET by id -> 404
            ret = client.get(
                f"{ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL}{ectbs_1_id}"
            )
            assert ret.status_code == 404

    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaigns")
    @pytest.mark.usefixtures("user_groups_by_campaign_scopes")
    @pytest.mark.parametrize("timeseries", (4,), indirect=True)
    def test_energy_consumption_timeseries_by_buildings_as_user_api(
        self,
        app,
        users,
        buildings,
        timeseries,
        energy_sources,
        energy_end_uses,
        energy_consumption_timeseries_by_buildings,
    ):

        user_creds = users["Active"]["creds"]
        building_1_id = buildings[0]
        ts_1_id = timeseries[0]
        ts_3_id = timeseries[2]
        energy_source_1_id = energy_sources[1]
        energy_end_use_1_id = energy_end_uses[1]
        ectbs_1_id = energy_consumption_timeseries_by_buildings[0]

        client = app.test_client()

        with AuthHeader(user_creds):

            # GET list
            ret = client.get(ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            ectbs_1 = ret_val[0]
            assert ectbs_1.pop("id") == ectbs_1_id

            # POST
            ectbs_3 = {
                "building_id": building_1_id,
                "timeseries_id": ts_1_id,
                "source_id": energy_source_1_id,
                "end_use_id": energy_end_use_1_id,
                "wh_conversion_factor": 1,
            }
            ret = client.post(
                ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL, json=ectbs_3
            )
            assert ret.status_code == 403

            # GET by id
            ret = client.get(
                f"{ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL}{ectbs_1_id}"
            )
            assert ret.status_code == 200
            ectbs_1_etag = ret.headers["ETag"]

            # PUT
            ectbs_1["timeseries_id"] = ts_3_id
            ret = client.put(
                f"{ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL}{ectbs_1_id}",
                json=ectbs_1,
                headers={"If-Match": ectbs_1_etag},
            )
            assert ret.status_code == 403

            # DELETE
            ret = client.delete(
                f"{ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL}{ectbs_1_id}",
                headers={"If-Match": ectbs_1_etag},
            )
            assert ret.status_code == 403

    def test_energy_consumption_timeseries_by_buildings_as_anonym_api(
        self, app, buildings, timeseries, energy_consumption_timeseries_by_buildings
    ):
        building_1_id = buildings[0]
        ts_1_id = timeseries[0]
        ts_2_id = timeseries[1]
        ectbs_1_id = energy_consumption_timeseries_by_buildings[0]

        client = app.test_client()

        # GET list
        ret = client.get(ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL)
        assert ret.status_code == 401

        # POST
        ectbs_3 = {
            "building_id": building_1_id,
            "timeseries_id": ts_2_id,
        }
        ret = client.post(ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL, json=ectbs_3)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(
            f"{ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL}{ectbs_1_id}"
        )
        assert ret.status_code == 401

        # PUT
        ectbs_1 = {
            "building_id": building_1_id,
            "timeseries_id": ts_1_id,
        }
        ret = client.put(
            f"{ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL}{ectbs_1_id}",
            json=ectbs_1,
            headers={"If-Match": "Dummy-ETag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401

        # DELETE
        ret = client.delete(
            f"{ENERGY_CONSUMPTION_TIMESERIES_BY_BUILDINGS_URL}{ectbs_1_id}",
            headers={"If-Match": "Dummy-Etag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401



"""BEMServer API"""
import flask

from bemserver_core import BEMServerCore

from . import database
from .extensions import (  # noqa
    Api,
    Blueprint,
    Schema,
    AutoSchema,
    SQLCursorPage,
    SortField,
    authentication,
)
from .resources import register_blueprints


__version__ = "0.4.0"

API_VERSION = __version__
OPENAPI_VERSION = "3.1.0"


def create_app(config_override=None):
    """Create application

    :param type config_override: Config class overriding default config.
        Used for tests.
    """
    app = flask.Flask(__name__)
    app.config.from_object("bemserver_api.settings.Config")
    app.config.from_envvar("FLASK_SETTINGS_FILE", silent=True)
    app.config.from_object(config_override)

    database.init_app(app)
    api = Api(
        spec_kwargs={
            "version": API_VERSION,
            "openapi_version": OPENAPI_VERSION,
        }
    )
    api.init_app(app)
    register_blueprints(api)

    bsc = BEMServerCore()
    bsc.init_auth()

    return app



"""Timeseries property data tests"""
import copy
import pytest

from tests.common import AuthHeader


DUMMY_ID = "69"

TIMESERIES_PROPERTY_DATA_URL = "/timeseries_property_data/"


class TestTimeseriesPropertyDataApi:
    def test_timeseries_property_data_api(
        self, app, users, timeseries_properties, timeseries
    ):

        tsp_1_id = timeseries_properties[0]
        tsp_2_id = timeseries_properties[1]
        ts_1_id = timeseries[0]

        creds = users["Chuck"]["creds"]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(TIMESERIES_PROPERTY_DATA_URL)
            assert ret.status_code == 200
            assert ret.json == []

            # POST
            tsg_1 = {
                "timeseries_id": ts_1_id,
                "property_id": tsp_1_id,
                "value": "12",
            }
            ret = client.post(TIMESERIES_PROPERTY_DATA_URL, json=tsg_1)
            assert ret.status_code == 201
            ret_val = ret.json
            tsg_1_id = ret_val.pop("id")
            tsg_1_etag = ret.headers["ETag"]
            assert ret_val == tsg_1

            # POST violating unique constraint
            ret = client.post(TIMESERIES_PROPERTY_DATA_URL, json=tsg_1)
            assert ret.status_code == 409

            # POST wrong value type
            tsg_post = {
                "timeseries_id": ts_1_id,
                "property_id": tsp_2_id,
                "value": "wrong type",
            }
            ret = client.post(TIMESERIES_PROPERTY_DATA_URL, json=tsg_post)
            assert ret.status_code == 422

            # GET list
            ret = client.get(TIMESERIES_PROPERTY_DATA_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == tsg_1_id

            # GET by id
            ret = client.get(f"{TIMESERIES_PROPERTY_DATA_URL}{tsg_1_id}")
            assert ret.status_code == 200
            assert ret.headers["ETag"] == tsg_1_etag
            ret_val = ret.json
            ret_val.pop("id")
            assert ret_val == tsg_1

            # PUT
            tsg_1["value"] = "42"
            ret = client.put(
                f"{TIMESERIES_PROPERTY_DATA_URL}{tsg_1_id}",
                json=tsg_1,
                headers={"If-Match": tsg_1_etag},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            ret_val.pop("id")
            tsg_1_etag = ret.headers["ETag"]
            assert ret_val == tsg_1

            # PUT wrong value type
            tsg_1_put = copy.deepcopy(tsg_1)
            tsg_1_put["value"] = "wrong type"
            ret = client.put(
                f"{TIMESERIES_PROPERTY_DATA_URL}{tsg_1_id}",
                json=tsg_1_put,
                headers={"If-Match": tsg_1_etag},
            )
            assert ret.status_code == 422

            # PUT wrong ID -> 404
            ret = client.put(
                f"{TIMESERIES_PROPERTY_DATA_URL}{DUMMY_ID}",
                json=tsg_1,
                headers={"If-Match": tsg_1_etag},
            )
            assert ret.status_code == 404

            # POST TSPD 2
            tsg_2 = {
                "timeseries_id": ts_1_id,
                "property_id": tsp_2_id,
                "value": "42",
            }
            ret = client.post(TIMESERIES_PROPERTY_DATA_URL, json=tsg_2)
            ret_val = ret.json
            tsg_2_id = ret_val.pop("id")
            tsg_2_etag = ret.headers["ETag"]

            # PUT violating unique constraint
            tsg_2["property_id"] = tsp_1_id
            ret = client.put(
                f"{TIMESERIES_PROPERTY_DATA_URL}{tsg_2_id}",
                json=tsg_2,
                headers={"If-Match": tsg_2_etag},
            )
            assert ret.status_code == 409

            # DELETE
            ret = client.delete(
                f"{TIMESERIES_PROPERTY_DATA_URL}{tsg_1_id}",
                headers={"If-Match": tsg_1_etag},
            )
            assert ret.status_code == 204
            ret = client.delete(
                f"{TIMESERIES_PROPERTY_DATA_URL}{tsg_2_id}",
                headers={"If-Match": tsg_2_etag},
            )

            # GET list
            ret = client.get(TIMESERIES_PROPERTY_DATA_URL)
            assert ret.status_code == 200
            assert ret.json == []

            # GET by id -> 404
            ret = client.get(f"{TIMESERIES_PROPERTY_DATA_URL}{tsg_1_id}")
            assert ret.status_code == 404

    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaign_scopes")
    def test_timeseries_property_data_as_user_api(
        self,
        app,
        users,
        timeseries_properties,
        timeseries,
        timeseries_property_data,
    ):

        tsp_1_id = timeseries_properties[0]
        ts_1_id = timeseries[0]
        tspd_1_id = timeseries_property_data[0]
        tspd_3_id = timeseries_property_data[2]

        creds = users["Active"]["creds"]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(TIMESERIES_PROPERTY_DATA_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2
            assert all([tspd["timeseries_id"] == ts_1_id for tspd in ret_val])

            # POST
            tspd = {
                "timeseries_id": ts_1_id,
                "property_id": tsp_1_id,
                "value": "12",
            }
            ret = client.post(TIMESERIES_PROPERTY_DATA_URL, json=tspd)
            # This would trigger a unique constraint violation error
            # but we get rejected before that
            assert ret.status_code == 403

            # GET by id
            ret = client.get(f"{TIMESERIES_PROPERTY_DATA_URL}{tspd_1_id}")
            assert ret.status_code == 200
            ts_1_etag = ret.headers["ETag"]
            ret_val = ret.json
            ret_val.pop("id")
            tspd_1 = ret_val

            # GET by id, user not in group
            ret = client.get(f"{TIMESERIES_PROPERTY_DATA_URL}{tspd_3_id}")
            assert ret.status_code == 403

            # PUT
            tspd_1["value"] = "42"
            ret = client.put(
                f"{TIMESERIES_PROPERTY_DATA_URL}{tspd_1_id}",
                json=tspd_1,
                headers={"If-Match": ts_1_etag},
            )
            # ETag is wrong but we get rejected before ETag check anyway
            assert ret.status_code == 403

            # DELETE
            ret = client.delete(
                f"{TIMESERIES_PROPERTY_DATA_URL}{tspd_1_id}",
                headers={"If-Match": ts_1_etag},
            )
            # ETag is wrong but we get rejected before ETag check anyway
            assert ret.status_code == 403

    def test_timeseries_property_data_as_anonym_api(
        self,
        app,
        users,
        timeseries_properties,
        timeseries,
        timeseries_property_data,
    ):

        tsp_1_id = timeseries_properties[0]
        ts_1_id = timeseries[0]
        tspd_1_id = timeseries_property_data[0]

        client = app.test_client()

        # GET list
        ret = client.get(TIMESERIES_PROPERTY_DATA_URL)
        assert ret.status_code == 401

        # POST
        tspd = {
            "timeseries_id": ts_1_id,
            "property_id": tsp_1_id,
            "value": "12",
        }
        ret = client.post(TIMESERIES_PROPERTY_DATA_URL, json=tspd)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(f"{TIMESERIES_PROPERTY_DATA_URL}{tspd_1_id}")
        assert ret.status_code == 401

        # PUT
        tspd["value"] = "42"
        ret = client.put(
            f"{TIMESERIES_PROPERTY_DATA_URL}{tspd_1_id}",
            json=tspd,
            headers={"If-Match": "Dummy-ETag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401

        # DELETE
        ret = client.delete(
            f"{TIMESERIES_PROPERTY_DATA_URL}{tspd_1_id}",
            headers={"If-Match": "Dummy-ETag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401



"""Timeseries property data"""
import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.model import TimeseriesPropertyData

from bemserver_api import AutoSchema, Schema


class TimeseriesPropertyDataSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = TimeseriesPropertyData

    id = msa.auto_field(dump_only=True)


class TimeseriesPropertyDataQueryArgsSchema(Schema):
    timeseries_id = ma.fields.Int()
    property_id = ma.fields.Int()



"""Space property data API schemas"""

import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.model import SpacePropertyData

from bemserver_api import AutoSchema, Schema


class SpacePropertyDataSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = SpacePropertyData

    id = msa.auto_field(dump_only=True)


class SpacePropertyDataQueryArgsSchema(Schema):
    space_id = ma.fields.Int()
    space_property_id = ma.fields.Int()



"""Global conftest"""
import base64
import datetime as dt

import sqlalchemy as sqla
import flask.testing

from bemserver_core.database import db
from bemserver_core.authorization import OpenBar
from bemserver_core import common, model, scheduled_tasks
from bemserver_core.commands import setup_db

import pytest
from pytest_postgresql import factories as ppf

from bemserver_api import create_app

from tests.common import TestConfig, AUTH_HEADER


postgresql_proc = ppf.postgresql_proc(
    postgres_options=(
        "-c shared_preload_libraries='timescaledb' "
        "-c timescaledb.telemetry_level=off"
    )
)
postgresql = ppf.postgresql("postgresql_proc")


def _get_db_url(postgresql):
    return (
        "postgresql+psycopg2://"
        f"{postgresql.info.user}:{postgresql.info.password}"
        f"@{postgresql.info.host}:{postgresql.info.port}/"
        f"{postgresql.info.dbname}"
    )


@pytest.fixture
def postgresql_db(postgresql):
    yield _get_db_url(postgresql)


@pytest.fixture
def timescale_db(postgresql_db):
    with sqla.create_engine(postgresql_db).connect() as connection:
        connection.execute("CREATE EXTENSION IF NOT EXISTS timescaledb;")
    yield postgresql_db


class TestClient(flask.testing.FlaskClient):
    def open(self, *args, **kwargs):
        auth_header = AUTH_HEADER.get()
        if auth_header:
            (kwargs.setdefault("headers", {}).setdefault("Authorization", auth_header))
        return super().open(*args, **kwargs)


@pytest.fixture(params=(TestConfig,))
def app(request, timescale_db):
    class AppConfig(request.param):
        SQLALCHEMY_DATABASE_URI = timescale_db

    application = create_app(AppConfig)
    application.test_client_class = TestClient
    setup_db()
    yield application
    db.session.remove()


USERS = (
    ("Chuck", "N0rris", "chuck@test.com", True, True),
    ("Active", "@ctive", "active@test.com", False, True),
    ("Inactive", "in@ctive", "inactive@test.com", False, False),
)


@pytest.fixture(params=(USERS,))
def users(app, request):
    with OpenBar():
        ret = {}
        for user in request.param:
            name, password, email, is_admin, is_active = user
            user = model.User.new(
                name=name, email=email, is_admin=is_admin, is_active=is_active
            )
            user.set_password(password)
            creds = base64.b64encode(f"{email}:{password}".encode()).decode()
            invalid_creds = base64.b64encode(f"{email}:bad_pwd".encode()).decode()
            ret[name] = {"user": user, "creds": creds, "invalid_creds": invalid_creds}
        db.session.commit()
        # Set id after commit
        for user in ret.values():
            user["id"] = user["user"].id
    return ret


@pytest.fixture
def user_groups(app):
    with OpenBar():
        user_group_1 = model.UserGroup.new(
            name="User group 1",
        )
        user_group_2 = model.UserGroup.new(
            name="User group 2",
        )
        db.session.commit()
    return (user_group_1.id, user_group_2.id)


@pytest.fixture
def users_by_user_groups(app, users, user_groups):
    with OpenBar():
        ubug_1 = model.UserByUserGroup.new(
            user_id=users["Active"]["id"],
            user_group_id=user_groups[0],
        )
        ubug_2 = model.UserByUserGroup.new(
            user_id=users["Inactive"]["id"],
            user_group_id=user_groups[1],
        )
        db.session.commit()
    return (ubug_1.id, ubug_2.id)


@pytest.fixture
def campaigns(app):
    with OpenBar():
        campaign_1 = model.Campaign.new(
            name="Campaign 1",
            start_time=dt.datetime(2020, 1, 1, tzinfo=dt.timezone.utc),
            timezone="UTC",
        )
        campaign_2 = model.Campaign.new(
            name="Campaign 2",
            start_time=dt.datetime(2020, 1, 1, tzinfo=dt.timezone.utc),
            end_time=dt.datetime(2021, 1, 1, tzinfo=dt.timezone.utc),
            timezone="Europe/Paris",
        )
        db.session.commit()
    return campaign_1.id, campaign_2.id


@pytest.fixture
def campaign_scopes(app, campaigns):
    with OpenBar():
        cs_1 = model.CampaignScope.new(
            name="Campaign 1 - Scope 1",
            campaign_id=campaigns[0],
        )
        cs_2 = model.CampaignScope.new(
            name="Campaign 2 - Scope 1",
            campaign_id=campaigns[1],
        )
        db.session.commit()
    return (cs_1.id, cs_2.id)


@pytest.fixture
def user_groups_by_campaigns(app, user_groups, campaigns):
    with OpenBar():
        ugbc_1 = model.UserGroupByCampaign.new(
            user_group_id=user_groups[0],
            campaign_id=campaigns[0],
        )
        ugbc_2 = model.UserGroupByCampaign.new(
            user_group_id=user_groups[1],
            campaign_id=campaigns[1],
        )
        db.session.commit()
    return (ugbc_1.id, ugbc_2.id)


@pytest.fixture
def user_groups_by_campaign_scopes(app, user_groups, campaign_scopes):
    with OpenBar():
        ugbcs_1 = model.UserGroupByCampaignScope.new(
            user_group_id=user_groups[0],
            campaign_scope_id=campaign_scopes[0],
        )
        ugbcs_2 = model.UserGroupByCampaignScope.new(
            user_group_id=user_groups[1],
            campaign_scope_id=campaign_scopes[1],
        )
        db.session.commit()
    return (ugbcs_1.id, ugbcs_2.id)


@pytest.fixture
def timeseries_properties(app):
    with OpenBar():
        return [x.id for x in model.TimeseriesProperty.get()]


@pytest.fixture(params=[2])
def timeseries(request, app, campaigns, campaign_scopes):
    with OpenBar():
        ts_l = []
        for i in range(request.param):
            ts_i = model.Timeseries(
                name=f"Timeseries {i}",
                description=f"Test timeseries #{i}",
                campaign_id=campaigns[i % len(campaigns)],
                campaign_scope_id=campaign_scopes[i % len(campaign_scopes)],
            )
            ts_l.append(ts_i)
        db.session.add_all(ts_l)
        db.session.commit()
        return [ts.id for ts in ts_l]


@pytest.fixture
def timeseries_property_data(request, app, timeseries_properties, timeseries):
    with OpenBar():
        tspd_l = []
        for ts in timeseries:
            tspd_l.append(
                model.TimeseriesPropertyData(
                    timeseries_id=ts,
                    property_id=timeseries_properties[0],
                    value=12,
                )
            )
            tspd_l.append(
                model.TimeseriesPropertyData(
                    timeseries_id=ts,
                    property_id=timeseries_properties[1],
                    value=42,
                )
            )
        db.session.add_all(tspd_l)
        db.session.commit()
        return [tspd.id for tspd in tspd_l]


@pytest.fixture(params=[2])
def timeseries_by_data_states(request, app, timeseries):
    with OpenBar():
        ts_l = []
        for i in range(request.param):
            ts_i = model.TimeseriesByDataState(
                timeseries_id=timeseries[i % len(timeseries)],
                data_state_id=1,
            )
            ts_l.append(ts_i)
        db.session.add_all(ts_l)
        db.session.commit()
        return [ts.id for ts in ts_l]


@pytest.fixture(params=[4])
def timeseries_data(request, app, timeseries_by_data_states):
    with OpenBar():
        nb_tsd = request.param
        for tsbds_id in timeseries_by_data_states:
            start_dt = dt.datetime(2020, 1, 1, tzinfo=dt.timezone.utc)
            for i in range(nb_tsd):
                timestamp = start_dt + dt.timedelta(hours=i)
                model.TimeseriesData.new(
                    timestamp=timestamp, timeseries_by_data_state_id=tsbds_id, value=i
                )
        db.session.commit()
    return (start_dt, start_dt + dt.timedelta(hours=nb_tsd))


@pytest.fixture
def event_categories(app):
    with OpenBar():
        ec_1 = model.EventCategory.new(name="Custom event category 1")
        ec_2 = model.EventCategory.new(name="Custom event category 2")
        db.session.commit()
    return (ec_1.id, ec_2.id)


@pytest.fixture
def event_levels(app):
    with OpenBar():
        el_1 = model.EventLevel.get(name="WARNING").first()
        el_2 = model.EventLevel.get(name="INFO").first()
        db.session.commit()
    return (el_1.id, el_2.id)


@pytest.fixture
def events(app, campaigns, campaign_scopes, event_categories, event_levels):
    with OpenBar():
        tse_1 = model.Event.new(
            campaign_scope_id=campaign_scopes[0],
            timestamp=dt.datetime(2020, 1, 1, tzinfo=dt.timezone.utc),
            source="Event source",
            category_id=event_categories[0],
            level_id=event_levels[0],
        )
        tse_2 = model.Event.new(
            campaign_scope_id=campaign_scopes[1],
            timestamp=dt.datetime(2021, 1, 1, tzinfo=dt.timezone.utc),
            source="Another event source",
            category_id=event_categories[1],
            level_id=event_levels[1],
        )
        db.session.commit()
    return (tse_1.id, tse_2.id)


@pytest.fixture
def timeseries_by_events(app, events, timeseries):
    with OpenBar():
        tbs_1 = model.TimeseriesByEvent.new(
            event_id=events[0],
            timeseries_id=timeseries[0],
        )
        tbs_2 = model.TimeseriesByEvent.new(
            event_id=events[1],
            timeseries_id=timeseries[1],
        )
        db.session.commit()
    return (tbs_1.id, tbs_2.id)


@pytest.fixture
def events_by_sites(app, sites, events):
    with OpenBar():
        ebs_1 = model.EventBySite.new(
            event_id=events[0],
            site_id=sites[0],
        )
        ebs_2 = model.EventBySite.new(
            event_id=events[1],
            site_id=sites[1],
        )
        db.session.commit()
    return (ebs_1.id, ebs_2.id)


@pytest.fixture
def events_by_buildings(app, buildings, events):
    with OpenBar():
        ebb_1 = model.EventByBuilding.new(
            event_id=events[0],
            building_id=buildings[0],
        )
        ebb_2 = model.EventByBuilding.new(
            event_id=events[1],
            building_id=buildings[1],
        )
        db.session.commit()
    return (ebb_1.id, ebb_2.id)


@pytest.fixture
def events_by_storeys(app, storeys, events):
    with OpenBar():
        ebs_1 = model.EventByStorey.new(
            event_id=events[0],
            storey_id=storeys[0],
        )
        ebs_2 = model.EventByStorey.new(
            event_id=events[1],
            storey_id=storeys[1],
        )
        db.session.commit()
    return (ebs_1.id, ebs_2.id)


@pytest.fixture
def events_by_spaces(app, spaces, events):
    with OpenBar():
        ebs_1 = model.EventBySpace.new(
            event_id=events[0],
            space_id=spaces[0],
        )
        ebs_2 = model.EventBySpace.new(
            event_id=events[1],
            space_id=spaces[1],
        )
        db.session.commit()
    return (ebs_1.id, ebs_2.id)


@pytest.fixture
def events_by_zones(app, zones, events):
    with OpenBar():
        ebz_1 = model.EventByZone.new(
            event_id=events[0],
            zone_id=zones[0],
        )
        ebz_2 = model.EventByZone.new(
            event_id=events[1],
            zone_id=zones[1],
        )
        db.session.commit()
    return (ebz_1.id, ebz_2.id)


@pytest.fixture
def sites(app, campaigns):
    with OpenBar():
        site_1 = model.Site.new(
            name="Site 1",
            campaign_id=campaigns[0],
        )
        site_2 = model.Site.new(
            name="Site 2",
            campaign_id=campaigns[1],
        )
        db.session.commit()
    return (site_1.id, site_2.id)


@pytest.fixture
def buildings(app, sites):
    with OpenBar():
        building_1 = model.Building.new(
            name="Building 1",
            site_id=sites[0],
        )
        building_2 = model.Building.new(
            name="Building 2",
            site_id=sites[1],
        )
        db.session.commit()
    return (building_1.id, building_2.id)


@pytest.fixture
def storeys(app, buildings):
    with OpenBar():
        storey_1 = model.Storey.new(
            name="Storey 1",
            building_id=buildings[0],
        )
        storey_2 = model.Storey.new(
            name="Storey 2",
            building_id=buildings[1],
        )
        db.session.commit()
    return (storey_1.id, storey_2.id)


@pytest.fixture
def spaces(app, storeys):
    with OpenBar():
        space_1 = model.Space.new(
            name="Space 1",
            storey_id=storeys[0],
        )
        space_2 = model.Space.new(
            name="Space 2",
            storey_id=storeys[1],
        )
        db.session.commit()
    return (space_1.id, space_2.id)


@pytest.fixture
def zones(app, campaigns):
    with OpenBar():
        zone_1 = model.Zone.new(
            name="Zone 1",
            campaign_id=campaigns[0],
        )
        zone_2 = model.Zone.new(
            name="Zone 2",
            campaign_id=campaigns[1],
        )
        db.session.commit()
    return (zone_1.id, zone_2.id)


@pytest.fixture
def structural_element_properties(app):
    with OpenBar():
        sep_1 = model.StructuralElementProperty.new(
            name="Area",
            value_type=common.PropertyType.integer,
        )
        sep_2 = model.StructuralElementProperty.new(
            name="Volume",
            value_type=common.PropertyType.float,
        )
        sep_3 = model.StructuralElementProperty.new(
            name="Window state",
            value_type=common.PropertyType.boolean,
        )
        sep_4 = model.StructuralElementProperty.new(
            name="Architect",
            value_type=common.PropertyType.string,
        )
        db.session.commit()
    return (sep_1.id, sep_2.id, sep_3.id, sep_4.id)


@pytest.fixture
def site_properties(app, structural_element_properties):
    with OpenBar():
        site_p_1 = model.SiteProperty.new(
            structural_element_property_id=structural_element_properties[0],
        )
        site_p_2 = model.SiteProperty.new(
            structural_element_property_id=structural_element_properties[1],
        )
        db.session.commit()
    return (site_p_1.id, site_p_2.id)


@pytest.fixture
def building_properties(app, structural_element_properties):
    with OpenBar():
        building_p_1 = model.BuildingProperty.new(
            structural_element_property_id=structural_element_properties[0],
        )
        building_p_2 = model.BuildingProperty.new(
            structural_element_property_id=structural_element_properties[1],
        )
        db.session.commit()
    return (building_p_1.id, building_p_2.id)


@pytest.fixture
def storey_properties(app, structural_element_properties):
    with OpenBar():
        storey_p_1 = model.StoreyProperty.new(
            structural_element_property_id=structural_element_properties[0],
        )
        storey_p_2 = model.StoreyProperty.new(
            structural_element_property_id=structural_element_properties[1],
        )
        db.session.commit()
    return (storey_p_1.id, storey_p_2.id)


@pytest.fixture
def space_properties(app, structural_element_properties):
    with OpenBar():
        space_p_1 = model.SpaceProperty.new(
            structural_element_property_id=structural_element_properties[0],
        )
        space_p_2 = model.SpaceProperty.new(
            structural_element_property_id=structural_element_properties[1],
        )
        db.session.commit()
    return (space_p_1.id, space_p_2.id)


@pytest.fixture
def zone_properties(app, structural_element_properties):
    with OpenBar():
        zone_p_1 = model.ZoneProperty.new(
            structural_element_property_id=structural_element_properties[0],
        )
        zone_p_2 = model.ZoneProperty.new(
            structural_element_property_id=structural_element_properties[1],
        )
        db.session.commit()
    return (zone_p_1.id, zone_p_2.id)


@pytest.fixture
def site_property_data(app, sites, site_properties):
    with OpenBar():
        spd_1 = model.SitePropertyData.new(
            site_id=sites[0],
            site_property_id=site_properties[0],
            value="12",
        )
        spd_2 = model.SitePropertyData.new(
            site_id=sites[1],
            site_property_id=site_properties[1],
            value="42",
        )
        db.session.commit()
    return (spd_1.id, spd_2.id)


@pytest.fixture
def building_property_data(app, sites, site_properties):
    with OpenBar():
        bpd_1 = model.BuildingPropertyData.new(
            building_id=sites[0],
            building_property_id=site_properties[0],
            value="12",
        )
        bpd_2 = model.BuildingPropertyData.new(
            building_id=sites[1],
            building_property_id=site_properties[1],
            value="42",
        )
        db.session.commit()
    return (bpd_1.id, bpd_2.id)


@pytest.fixture
def storey_property_data(app, storeys, storey_properties):
    with OpenBar():
        spd_1 = model.StoreyPropertyData.new(
            storey_id=storeys[0],
            storey_property_id=storey_properties[0],
            value="12",
        )
        spd_2 = model.StoreyPropertyData.new(
            storey_id=storeys[1],
            storey_property_id=storey_properties[1],
            value="42",
        )
        db.session.commit()
    return (spd_1.id, spd_2.id)


@pytest.fixture
def space_property_data(app, spaces, space_properties):
    with OpenBar():
        spd_1 = model.SpacePropertyData.new(
            space_id=spaces[0],
            space_property_id=space_properties[0],
            value="12",
        )
        spd_2 = model.SpacePropertyData.new(
            space_id=spaces[1],
            space_property_id=space_properties[1],
            value="42",
        )
        db.session.commit()
    return (spd_1.id, spd_2.id)


@pytest.fixture
def zone_property_data(app, zones, zone_properties):
    with OpenBar():
        zpd_1 = model.ZonePropertyData.new(
            zone_id=zones[0],
            zone_property_id=zone_properties[0],
            value="12",
        )
        zpd_2 = model.ZonePropertyData.new(
            zone_id=zones[1],
            zone_property_id=zone_properties[1],
            value="42",
        )
        db.session.commit()
    return (zpd_1.id, zpd_2.id)


@pytest.fixture
def timeseries_by_sites(app, sites, timeseries):
    with OpenBar():
        tbs_1 = model.TimeseriesBySite.new(
            site_id=sites[0],
            timeseries_id=timeseries[0],
        )
        tbs_2 = model.TimeseriesBySite.new(
            site_id=sites[1],
            timeseries_id=timeseries[1],
        )
        db.session.commit()
    return (tbs_1.id, tbs_2.id)


@pytest.fixture
def timeseries_by_buildings(app, buildings, timeseries):
    with OpenBar():
        tbb_1 = model.TimeseriesByBuilding.new(
            building_id=buildings[0],
            timeseries_id=timeseries[0],
        )
        tbb_2 = model.TimeseriesByBuilding.new(
            building_id=buildings[1],
            timeseries_id=timeseries[1],
        )
        db.session.commit()
    return (tbb_1.id, tbb_2.id)


@pytest.fixture
def timeseries_by_storeys(app, storeys, timeseries):
    with OpenBar():
        tbs_1 = model.TimeseriesByStorey.new(
            storey_id=storeys[0],
            timeseries_id=timeseries[0],
        )
        tbs_2 = model.TimeseriesByStorey.new(
            storey_id=storeys[1],
            timeseries_id=timeseries[1],
        )
        db.session.commit()
    return (tbs_1.id, tbs_2.id)


@pytest.fixture
def timeseries_by_spaces(app, spaces, timeseries):
    with OpenBar():
        tbs_1 = model.TimeseriesBySpace.new(
            space_id=spaces[0],
            timeseries_id=timeseries[0],
        )
        tbs_2 = model.TimeseriesBySpace.new(
            space_id=spaces[1],
            timeseries_id=timeseries[1],
        )
        db.session.commit()
    return (tbs_1.id, tbs_2.id)


@pytest.fixture
def timeseries_by_zones(app, spaces, timeseries):
    with OpenBar():
        tbz_1 = model.TimeseriesByZone.new(
            zone_id=spaces[0],
            timeseries_id=timeseries[0],
        )
        tbz_2 = model.TimeseriesByZone.new(
            zone_id=spaces[1],
            timeseries_id=timeseries[1],
        )
        db.session.commit()
    return (tbz_1.id, tbz_2.id)


@pytest.fixture
def energy_sources(app):
    with OpenBar():
        return [x.id for x in model.EnergySource.get()]


@pytest.fixture
def energy_end_uses(app):
    with OpenBar():
        return [x.id for x in model.EnergyEndUse.get()]


@pytest.fixture
def energy_consumption_timeseries_by_sites(app, timeseries, sites):
    with OpenBar():
        ectbs_1 = model.EnergyConsumptionTimeseriesBySite.new(
            site_id=sites[0],
            source_id=1,
            end_use_id=1,
            timeseries_id=timeseries[0],
        )
        ectbs_2 = model.EnergyConsumptionTimeseriesBySite.new(
            site_id=sites[1],
            source_id=2,
            end_use_id=2,
            timeseries_id=timeseries[1],
        )
        db.session.commit()
    return (ectbs_1.id, ectbs_2.id)


@pytest.fixture
def energy_consumption_timeseries_by_buildings(app, timeseries, buildings):
    with OpenBar():
        ectbs_1 = model.EnergyConsumptionTimeseriesByBuilding.new(
            building_id=buildings[0],
            source_id=1,
            end_use_id=1,
            timeseries_id=timeseries[0],
        )
        ectbs_2 = model.EnergyConsumptionTimeseriesByBuilding.new(
            building_id=buildings[1],
            source_id=2,
            end_use_id=2,
            timeseries_id=timeseries[1],
        )
        db.session.commit()
    return (ectbs_1.id, ectbs_2.id)


@pytest.fixture
def st_cleanups_by_campaigns(app, campaigns):
    with OpenBar():
        st_cbc_1 = scheduled_tasks.ST_CleanupByCampaign.new(campaign_id=campaigns[0])
        st_cbc_2 = scheduled_tasks.ST_CleanupByCampaign.new(campaign_id=campaigns[1])
        db.session.commit()
    return (st_cbc_1.id, st_cbc_2.id)


@pytest.fixture
def st_cleanups_by_timeseries(app, st_cleanups_by_campaigns, timeseries):
    with OpenBar():
        st_cbt_1 = scheduled_tasks.ST_CleanupByTimeseries.new(
            timeseries_id=timeseries[0]
        )
        st_cbt_2 = scheduled_tasks.ST_CleanupByTimeseries.new(
            timeseries_id=timeseries[1],
            last_timestamp=dt.datetime(2020, 1, 1, tzinfo=dt.timezone.utc),
        )
        db.session.commit()
    return (st_cbt_1.id, st_cbt_2.id)


@pytest.fixture
def st_check_missings_by_campaigns(app, campaigns):
    with OpenBar():
        st_cbc_1 = scheduled_tasks.ST_CheckMissingByCampaign.new(
            campaign_id=campaigns[0]
        )
        st_cbc_2 = scheduled_tasks.ST_CheckMissingByCampaign.new(
            campaign_id=campaigns[1]
        )
        db.session.commit()
    return (st_cbc_1.id, st_cbc_2.id)



"""Timeseries data API schemas"""
import marshmallow as ma

from bemserver_core.input_output.timeseries_data_io import AGGREGATION_FUNCTIONS
from bemserver_core.time_utils import PERIODS, FIXED_SIZE_PERIODS

from bemserver_api import Schema
from bemserver_api.extensions.ma_fields import Timezone


class TimeseriesIDListMixinSchema(Schema):
    timeseries = ma.fields.List(
        ma.fields.Int(),
        required=True,
        metadata={
            "description": "List of timeseries ID",
        },
    )


class TimeseriesNameListMixinSchema(Schema):
    timeseries = ma.fields.List(
        ma.fields.String(),
        required=True,
        metadata={
            "description": "List of timeseries names",
        },
    )


class TimeseriesDataBaseQueryArgsSchema(Schema):
    """Timeseries values query parameters base schema"""

    start_time = ma.fields.AwareDateTime(
        required=True,
        metadata={
            "description": "Initial datetime",
        },
    )
    end_time = ma.fields.AwareDateTime(
        required=True,
        metadata={
            "description": "End datetime (excluded from the interval)",
        },
    )
    data_state = ma.fields.Int(
        required=True,
        metadata={
            "description": "Data state ID",
        },
    )


class TimeseriesDataDeleteByIDQueryArgsSchema(
    TimeseriesDataBaseQueryArgsSchema, TimeseriesIDListMixinSchema
):
    """Timeseries values DELETE by ID query parameters schema"""


class TimeseriesDataDeleteByNameQueryArgsSchema(
    TimeseriesDataBaseQueryArgsSchema, TimeseriesNameListMixinSchema
):
    """Timeseries values DELETE by name query parameters schema"""


class TimeseriesDataGetBaseQueryArgsSchema(TimeseriesDataBaseQueryArgsSchema):
    """Timeseries values GET query parameters base schema"""

    timezone = Timezone(
        load_default="UTC",
        metadata={
            "description": "Timezone to use for response data",
        },
    )


class TimeseriesDataGetByIDQueryArgsSchema(
    TimeseriesDataGetBaseQueryArgsSchema, TimeseriesIDListMixinSchema
):
    """Timeseries values GET by ID query parameters schema"""


class TimeseriesDataGetByNameQueryArgsSchema(
    TimeseriesDataGetBaseQueryArgsSchema, TimeseriesNameListMixinSchema
):
    """Timeseries values GET by name query parameters schema"""


class TimeseriesBucketWidthSchema(Schema):
    """Timeseries bucket width schema, with custom validation on fixed size periods"""

    _VALIDATION_MSG = (
        f"Only fixed bucket width units ({', '.join(FIXED_SIZE_PERIODS)})"
        " accept a value different than 1."
    )

    bucket_width_value = ma.fields.Int(
        validate=ma.validate.Range(min=1),
        required=True,
        metadata={
            "description": _VALIDATION_MSG,
        },
    )
    bucket_width_unit = ma.fields.String(
        validate=ma.validate.OneOf(PERIODS),
        required=True,
    )

    @ma.validates_schema
    def validate_bucket(self, data, **kwargs):
        if (
            data["bucket_width_unit"] not in FIXED_SIZE_PERIODS
            and data["bucket_width_value"] != 1
        ):
            raise ma.ValidationError(
                self._VALIDATION_MSG,
                field_name="bucket_width_value",
                data=data["bucket_width_value"],
                valid_data=FIXED_SIZE_PERIODS,
            )


class TimeseriesDataGetAggregateBaseQueryArgsSchema(
    TimeseriesDataGetBaseQueryArgsSchema,
    TimeseriesBucketWidthSchema,
):
    """Timeseries values aggregate GET query parameters base schema"""

    aggregation = ma.fields.String(
        load_default="avg",
        validate=ma.validate.OneOf(AGGREGATION_FUNCTIONS),
    )


class TimeseriesDataGetByIDAggregateQueryArgsSchema(
    TimeseriesDataGetAggregateBaseQueryArgsSchema, TimeseriesIDListMixinSchema
):
    """Timeseries values aggregate GET by ID query parameters schema"""


class TimeseriesDataGetByNameAggregateQueryArgsSchema(
    TimeseriesDataGetAggregateBaseQueryArgsSchema, TimeseriesNameListMixinSchema
):
    """Timeseries values aggregate GET by name query parameters schema"""


class TimeseriesDataPostQueryArgsSchema(Schema):
    """Timeseries values POST query parameters schema"""

    data_state = ma.fields.Int(
        required=True,
        metadata={
            "description": "Data state ID",
        },
    )


class TimeseriesDataGetAcceptHeader(Schema):
    accept = ma.fields.String(
        validate=ma.validate.OneOf(["application/json", "application/csv"]),
        data_key="Accept",
        load_default="application/json",
    )



"""Events resources"""

from flask.views import MethodView
from flask_smorest import abort

from bemserver_core.model import Event

from bemserver_api import Blueprint, SQLCursorPage
from bemserver_api.database import db

from .schemas import (
    EventSchema,
    EventPutSchema,
    EventQueryArgsSchema,
)


blp = Blueprint(
    "Event", __name__, url_prefix="/events", description="Operations on events"
)


@blp.route("/")
class EventsViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.arguments(EventQueryArgsSchema, location="query")
    @blp.response(200, EventSchema(many=True))
    @blp.paginate(SQLCursorPage)
    def get(self, args):
        """List events"""
        return Event.get(**args)

    @blp.login_required
    @blp.etag
    @blp.arguments(EventSchema)
    @blp.response(201, EventSchema)
    @blp.catch_integrity_error
    def post(self, new_item):
        """Add a new event"""
        item = Event.new(**new_item)
        db.session.commit()
        return item


@blp.route("/<int:item_id>")
class EventsByIdViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.response(200, EventSchema)
    def get(self, item_id):
        """Get event by ID"""
        item = Event.get_by_id(item_id)
        if item is None:
            abort(404)
        return item

    @blp.login_required
    @blp.etag
    @blp.arguments(EventPutSchema)
    @blp.response(200, EventSchema)
    @blp.catch_integrity_error
    def put(self, new_item, item_id):
        """Update an existing event"""
        item = Event.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, EventSchema)
        item.update(**new_item)
        db.session.commit()
        return item

    @blp.login_required
    @blp.etag
    @blp.response(204)
    def delete(self, item_id):
        """Delete an event"""
        item = Event.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, EventSchema)
        item.delete()
        db.session.commit()



"""Energy tests"""
import pytest

from tests.common import AuthHeader


DUMMY_ID = "69"

ENERGY_SOURCES_URL = "/energy_sources/"
ENERGY_END_USES_URL = "/energy_end_uses/"


class TestEnergySourcesApi:
    @pytest.mark.parametrize("user", ("user", "admin"))
    def test_energy_sources_api_as_admin_or_user(self, app, user, users):

        if user == "user":
            creds = users["Active"]["creds"]
        else:
            creds = users["Chuck"]["creds"]

        client = app.test_client()

        with AuthHeader(creds):
            ret = client.get(ENERGY_SOURCES_URL)
            assert ret.status_code == 200
            assert "all" in [e["name"] for e in ret.json]

    def test_enery_sources_api_as_anonym(self, app):

        client = app.test_client()

        ret = client.get(ENERGY_SOURCES_URL)
        assert ret.status_code == 401


class TestEnergyEndUsesApi:
    @pytest.mark.parametrize("user", ("user", "admin"))
    def test_energy_end_uses_api_as_admin_or_user(self, app, user, users):

        if user == "user":
            creds = users["Active"]["creds"]
        else:
            creds = users["Chuck"]["creds"]

        client = app.test_client()

        with AuthHeader(creds):
            ret = client.get(ENERGY_END_USES_URL)
            assert ret.status_code == 200
            assert "all" in [e["name"] for e in ret.json]

    def test_enery_end_uses_api_as_anonym(self, app):

        client = app.test_client()

        ret = client.get(ENERGY_END_USES_URL)
        assert ret.status_code == 401



"""STCleanupByTimeseriess routes tests"""
import pytest

from tests.common import AuthHeader


DUMMY_ID = "69"

ST_CLEANUPS_BY_TIMESERIES_URL = "/st_cleanups_by_timeseries/"


class TestST_CleanupByTimeseriessApi:
    @pytest.mark.parametrize("timeseries", (10,), indirect=True)
    def test_st_cleanups_by_timeseries_api(
        self, app, users, campaigns, timeseries, st_cleanups_by_timeseries
    ):

        creds = users["Chuck"]["creds"]
        campaign_1_id = campaigns[0]
        ts_1_id = timeseries[0]
        ts_2_id = timeseries[1]
        ts_7_id = timeseries[6]
        st_cbt_1_id = st_cleanups_by_timeseries[0]
        st_cbt_2_id = st_cleanups_by_timeseries[1]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(ST_CLEANUPS_BY_TIMESERIES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2

            # GET list, filtered
            ret = client.get(
                ST_CLEANUPS_BY_TIMESERIES_URL,
                query_string={"campaign_id": campaign_1_id},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == st_cbt_1_id

            # GET by id
            ret = client.get(f"{ST_CLEANUPS_BY_TIMESERIES_URL}{st_cbt_1_id}")
            assert ret.status_code == 200
            ret_val = ret.json
            assert ret_val["id"] == st_cbt_1_id

            # GET list (full), sort by timeseries name
            ret = client.get(
                f"{ST_CLEANUPS_BY_TIMESERIES_URL}full",
                query_string={"sort": "+timeseries_name"},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 10
            assert ret_val[0]["id"] == st_cbt_1_id
            assert ret_val[0]["timeseries_id"] == ts_1_id
            assert ret_val[0]["last_timestamp"] is None

            # GET list (full), sort by timeseries name and timestamps desc
            ret = client.get(
                f"{ST_CLEANUPS_BY_TIMESERIES_URL}full",
                query_string={"sort": "-last_timestamp,+timeseries_name"},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 10
            assert ret_val[0]["id"] == st_cbt_2_id
            assert ret_val[0]["timeseries_id"] == ts_2_id
            assert ret_val[0]["last_timestamp"] == "2020-01-01T00:00:00+00:00"
            assert ret_val[1]["id"] == st_cbt_1_id
            assert ret_val[1]["timeseries_id"] == ts_1_id
            assert ret_val[1]["last_timestamp"] is None
            assert ret_val[6]["id"] is None
            assert ret_val[6]["timeseries_id"] == ts_7_id
            assert ret_val[6]["last_timestamp"] is None

    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaigns")
    @pytest.mark.usefixtures("user_groups_by_campaign_scopes")
    def test_st_cleanups_by_timeseries_as_user_api(
        self, app, users, campaigns, timeseries, st_cleanups_by_timeseries
    ):

        user_creds = users["Active"]["creds"]
        campaign_1_id = campaigns[0]
        campaign_2_id = campaigns[1]
        ts_1_id = timeseries[0]
        st_cbt_1_id = st_cleanups_by_timeseries[0]
        st_cbt_2_id = st_cleanups_by_timeseries[1]

        client = app.test_client()

        with AuthHeader(user_creds):

            # GET list
            ret = client.get(ST_CLEANUPS_BY_TIMESERIES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            st_cbt_1 = ret_val[0]
            assert st_cbt_1["id"] == st_cbt_1_id

            # GET list, filtered
            ret = client.get(
                ST_CLEANUPS_BY_TIMESERIES_URL,
                query_string={"campaign_id": campaign_1_id},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == st_cbt_1_id
            ret = client.get(
                ST_CLEANUPS_BY_TIMESERIES_URL,
                query_string={"campaign_id": campaign_2_id},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert not ret_val

            # GET by id
            ret = client.get(f"{ST_CLEANUPS_BY_TIMESERIES_URL}{st_cbt_1_id}")
            assert ret.status_code == 200
            ret_val = ret.json
            assert ret_val["id"] == st_cbt_1_id
            ret = client.get(f"{ST_CLEANUPS_BY_TIMESERIES_URL}{st_cbt_2_id}")
            assert ret.status_code == 403

            # GET list (full), sort by timeseries name
            ret = client.get(
                f"{ST_CLEANUPS_BY_TIMESERIES_URL}full",
                query_string={"sort": "+timeseries_name"},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == st_cbt_1_id
            assert ret_val[0]["timeseries_id"] == ts_1_id
            assert ret_val[0]["last_timestamp"] is None

            # GET list (full), filter by campaign id
            ret = client.get(
                f"{ST_CLEANUPS_BY_TIMESERIES_URL}full",
                query_string={"campaign_id": campaign_1_id},
            )
            assert ret.status_code == 200
            assert len(ret.json) == 1

            # GET list (full), filter by campaign id
            ret = client.get(
                f"{ST_CLEANUPS_BY_TIMESERIES_URL}full",
                query_string={"campaign_id": campaign_2_id},
            )
            assert ret.status_code == 403

    def test_st_cleanups_by_timeseries_as_anonym_api(
        self, app, st_cleanups_by_timeseries
    ):

        st_cbt_1_id = st_cleanups_by_timeseries[0]

        client = app.test_client()

        # GET list
        ret = client.get(ST_CLEANUPS_BY_TIMESERIES_URL)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(f"{ST_CLEANUPS_BY_TIMESERIES_URL}{st_cbt_1_id}")
        assert ret.status_code == 401

        # GET list (full)
        ret = client.get(f"{ST_CLEANUPS_BY_TIMESERIES_URL}full")
        assert ret.status_code == 401



"""Site property data API schemas"""

import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.model import SitePropertyData

from bemserver_api import AutoSchema, Schema


class SitePropertyDataSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = SitePropertyData

    id = msa.auto_field(dump_only=True)


class SitePropertyDataQueryArgsSchema(Schema):
    site_id = ma.fields.Int()
    site_property_id = ma.fields.Int()



"""Space property data routes tests"""
import copy
import pytest

from tests.common import AuthHeader


DUMMY_ID = "69"

SPACE_PROPERTY_DATA_URL = "/space_property_data/"
SPACES_URL = "/spaces/"


class TestSpacePropertyDataApi:
    def test_space_property_data_api(self, app, users, spaces, space_properties):

        creds = users["Chuck"]["creds"]
        space_1_id = spaces[0]
        space_2_id = spaces[1]
        space_p_1_id = space_properties[0]
        space_p_2_id = space_properties[1]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(SPACE_PROPERTY_DATA_URL)
            assert ret.status_code == 200
            assert ret.json == []

            # POST
            spd_1 = {
                "space_id": space_1_id,
                "space_property_id": space_p_1_id,
                "value": "12",
            }
            ret = client.post(SPACE_PROPERTY_DATA_URL, json=spd_1)
            assert ret.status_code == 201
            ret_val = ret.json
            spd_1_id = ret_val.pop("id")
            spd_1_etag = ret.headers["ETag"]
            assert ret_val == spd_1

            # POST violating unique constraint
            ret = client.post(SPACE_PROPERTY_DATA_URL, json=spd_1)
            assert ret.status_code == 409

            # POST wrong value type
            spd_post = {
                "space_id": space_1_id,
                "space_property_id": space_p_2_id,
                "value": "wrong type",
            }
            ret = client.post(SPACE_PROPERTY_DATA_URL, json=spd_post)
            assert ret.status_code == 422

            # GET list
            ret = client.get(SPACE_PROPERTY_DATA_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == space_p_1_id

            # GET by id
            ret = client.get(f"{SPACE_PROPERTY_DATA_URL}{spd_1_id}")
            assert ret.status_code == 200
            assert ret.headers["ETag"] == spd_1_etag
            ret_val = ret.json
            ret_val.pop("id")
            assert ret_val == spd_1

            # PUT
            spd_1["value"] = "69"
            ret = client.put(
                f"{SPACE_PROPERTY_DATA_URL}{spd_1_id}",
                json=spd_1,
                headers={"If-Match": spd_1_etag},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            ret_val.pop("id")
            spd_1_etag = ret.headers["ETag"]
            assert ret_val == spd_1

            # PUT wrong value type
            spd_1_put = copy.deepcopy(spd_1)
            spd_1_put["value"] = "wrong type"
            ret = client.put(
                f"{SPACE_PROPERTY_DATA_URL}{spd_1_id}",
                json=spd_1_put,
                headers={"If-Match": spd_1_etag},
            )
            assert ret.status_code == 422

            # PUT wrong ID -> 404
            ret = client.put(
                f"{SPACE_PROPERTY_DATA_URL}{DUMMY_ID}",
                json=spd_1,
                headers={"If-Match": spd_1_etag},
            )
            assert ret.status_code == 404

            # POST sep 2
            spd_2 = {
                "space_id": space_2_id,
                "space_property_id": space_p_2_id,
                "value": "42",
            }
            ret = client.post(SPACE_PROPERTY_DATA_URL, json=spd_2)
            ret_val = ret.json
            spd_2_id = ret_val.pop("id")
            spd_2_etag = ret.headers["ETag"]

            # PUT violating unique constraint
            spd_1["space_id"] = spd_2["space_id"]
            spd_1["space_property_id"] = spd_2["space_property_id"]
            ret = client.put(
                f"{SPACE_PROPERTY_DATA_URL}{spd_1_id}",
                json=spd_1,
                headers={"If-Match": spd_1_etag},
            )
            assert ret.status_code == 409

            # GET list
            ret = client.get(SPACE_PROPERTY_DATA_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2

            # GET list with filters
            ret = client.get(
                SPACE_PROPERTY_DATA_URL,
                query_string={"space_id": space_1_id},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == spd_1_id

            # DELETE wrong ID -> 404
            ret = client.delete(
                f"{SPACE_PROPERTY_DATA_URL}{DUMMY_ID}",
                headers={"If-Match": spd_1_etag},
            )
            assert ret.status_code == 404

            # DELETE space cascade
            ret = client.get(f"{SPACES_URL}{space_1_id}")
            space_1_etag = ret.headers["ETag"]
            ret = client.delete(
                f"{SPACES_URL}{space_1_id}", headers={"If-Match": space_1_etag}
            )
            assert ret.status_code == 204

            # DELETE
            ret = client.delete(
                f"{SPACE_PROPERTY_DATA_URL}{spd_1_id}",
                headers={"If-Match": spd_1_etag},
            )
            assert ret.status_code == 404
            ret = client.delete(
                f"{SPACE_PROPERTY_DATA_URL}{spd_2_id}",
                headers={"If-Match": spd_2_etag},
            )
            assert ret.status_code == 204

            # GET list
            ret = client.get(SPACE_PROPERTY_DATA_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # GET by id -> 404
            ret = client.get(f"{SPACE_PROPERTY_DATA_URL}{spd_1_id}")
            assert ret.status_code == 404

    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaigns")
    def test_space_property_data_as_user_api(
        self, app, users, spaces, space_properties, space_property_data
    ):

        user_creds = users["Active"]["creds"]
        space_1_id = spaces[0]
        space_p_1_id = space_properties[0]
        space_p_1_id = space_properties[0]
        spd_1_id = space_property_data[0]

        client = app.test_client()

        with AuthHeader(user_creds):

            # GET list
            ret = client.get(SPACE_PROPERTY_DATA_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            spd_1 = ret_val[0]
            assert spd_1.pop("id") == spd_1_id

            # POST
            spd_3 = {
                "space_id": space_1_id,
                "space_property_id": space_p_1_id,
                "value": "12",
            }
            ret = client.post(SPACE_PROPERTY_DATA_URL, json=spd_3)
            assert ret.status_code == 403

            # GET by id
            ret = client.get(f"{SPACE_PROPERTY_DATA_URL}{spd_1_id}")
            assert ret.status_code == 200
            spd_1_etag = ret.headers["ETag"]

            # PUT
            spd_1["value"] = "69"
            ret = client.put(
                f"{SPACE_PROPERTY_DATA_URL}{spd_1_id}",
                json=spd_1,
                headers={"If-Match": spd_1_etag},
            )
            assert ret.status_code == 403

            # DELETE
            ret = client.delete(
                f"{SPACE_PROPERTY_DATA_URL}{spd_1_id}",
                headers={"If-Match": spd_1_etag},
            )
            assert ret.status_code == 403

    def test_space_property_data_as_anonym_api(
        self, app, spaces, space_properties, space_property_data
    ):
        space_1_id = spaces[0]
        space_p_1_id = space_properties[0]
        space_p_2_id = space_properties[1]
        spd_1_id = space_property_data[0]

        client = app.test_client()

        # GET list
        ret = client.get(SPACE_PROPERTY_DATA_URL)
        assert ret.status_code == 401

        # POST
        spd_3 = {
            "space_id": space_1_id,
            "space_property_id": space_p_2_id,
            "value": "12",
        }
        ret = client.post(SPACE_PROPERTY_DATA_URL, json=spd_3)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(f"{SPACE_PROPERTY_DATA_URL}{spd_1_id}")
        assert ret.status_code == 401

        # PUT
        spd_1 = {
            "space_id": space_1_id,
            "space_property_id": space_p_1_id,
            "value": "12",
        }
        ret = client.put(
            f"{SPACE_PROPERTY_DATA_URL}{spd_1_id}",
            json=spd_1,
            headers={"If-Match": "Dummy-ETag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401

        # DELETE
        ret = client.delete(
            f"{SPACE_PROPERTY_DATA_URL}{spd_1_id}",
            headers={"If-Match": "Dummy-Etag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401



"""Timeseries by sites routes tests"""
import pytest

from tests.common import AuthHeader


DUMMY_ID = "69"

TIMESERIES_BY_SITES_URL = "/timeseries_by_sites/"
SITES_URL = "/sites/"


class TestTimeseriesBySiteApi:
    def test_timeseries_by_sites_api(self, app, users, sites, timeseries):

        creds = users["Chuck"]["creds"]
        site_1_id = sites[0]
        site_2_id = sites[1]
        ts_1_id = timeseries[0]
        ts_2_id = timeseries[1]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(TIMESERIES_BY_SITES_URL)
            assert ret.status_code == 200
            assert ret.json == []

            # POST
            tbs_1 = {
                "site_id": site_1_id,
                "timeseries_id": ts_1_id,
            }
            ret = client.post(TIMESERIES_BY_SITES_URL, json=tbs_1)
            assert ret.status_code == 201
            ret_val = ret.json
            tbs_1_id = ret_val.pop("id")
            tbs_1_etag = ret.headers["ETag"]
            assert ret_val == tbs_1

            # POST violating unique constraint
            ret = client.post(TIMESERIES_BY_SITES_URL, json=tbs_1)
            assert ret.status_code == 409

            # GET list
            ret = client.get(TIMESERIES_BY_SITES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == ts_1_id

            # GET by id
            ret = client.get(f"{TIMESERIES_BY_SITES_URL}{tbs_1_id}")
            assert ret.status_code == 200
            assert ret.headers["ETag"] == tbs_1_etag
            ret_val = ret.json
            ret_val.pop("id")
            assert ret_val == tbs_1

            # PUT
            tbs_1["timeseries_id"] = ts_2_id
            ret = client.put(
                f"{TIMESERIES_BY_SITES_URL}{tbs_1_id}",
                json=tbs_1,
                headers={"If-Match": tbs_1_etag},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            ret_val.pop("id")
            tbs_1_etag = ret.headers["ETag"]
            assert ret_val == tbs_1

            # PUT wrong ID -> 404
            ret = client.put(
                f"{TIMESERIES_BY_SITES_URL}{DUMMY_ID}",
                json=tbs_1,
                headers={"If-Match": tbs_1_etag},
            )
            assert ret.status_code == 404

            # POST sep 2
            tbs_2 = {
                "site_id": site_2_id,
                "timeseries_id": ts_2_id,
            }
            ret = client.post(TIMESERIES_BY_SITES_URL, json=tbs_2)
            ret_val = ret.json
            tbs_2_id = ret_val.pop("id")
            tbs_2_etag = ret.headers["ETag"]

            # PUT violating unique constraint
            tbs_1["site_id"] = tbs_2["site_id"]
            tbs_1["timeseries_id"] = tbs_2["timeseries_id"]
            ret = client.put(
                f"{TIMESERIES_BY_SITES_URL}{tbs_1_id}",
                json=tbs_1,
                headers={"If-Match": tbs_1_etag},
            )
            assert ret.status_code == 409

            # GET list
            ret = client.get(TIMESERIES_BY_SITES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2

            # GET list with filters
            ret = client.get(
                TIMESERIES_BY_SITES_URL,
                query_string={"site_id": site_1_id},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == tbs_1_id

            # DELETE wrong ID -> 404
            ret = client.delete(
                f"{TIMESERIES_BY_SITES_URL}{DUMMY_ID}",
                headers={"If-Match": tbs_1_etag},
            )
            assert ret.status_code == 404

            # DELETE site cascade
            ret = client.get(f"{SITES_URL}{site_1_id}")
            site_1_etag = ret.headers["ETag"]
            ret = client.delete(
                f"{SITES_URL}{site_1_id}", headers={"If-Match": site_1_etag}
            )
            assert ret.status_code == 204

            # DELETE
            ret = client.delete(
                f"{TIMESERIES_BY_SITES_URL}{tbs_1_id}",
                headers={"If-Match": tbs_1_etag},
            )
            assert ret.status_code == 404
            ret = client.delete(
                f"{TIMESERIES_BY_SITES_URL}{tbs_2_id}",
                headers={"If-Match": tbs_2_etag},
            )
            assert ret.status_code == 204

            # GET list
            ret = client.get(TIMESERIES_BY_SITES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # GET by id -> 404
            ret = client.get(f"{TIMESERIES_BY_SITES_URL}{tbs_1_id}")
            assert ret.status_code == 404

    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaigns")
    @pytest.mark.usefixtures("user_groups_by_campaign_scopes")
    @pytest.mark.parametrize("timeseries", (4,), indirect=True)
    def test_timeseries_by_sites_as_user_api(
        self, app, users, sites, timeseries, timeseries_by_sites
    ):

        user_creds = users["Active"]["creds"]
        site_1_id = sites[0]
        ts_1_id = timeseries[0]
        ts_3_id = timeseries[2]
        tbs_1_id = timeseries_by_sites[0]

        client = app.test_client()

        with AuthHeader(user_creds):

            # GET list
            ret = client.get(TIMESERIES_BY_SITES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            tbs_1 = ret_val[0]
            assert tbs_1.pop("id") == tbs_1_id

            # POST
            tbs_3 = {
                "site_id": site_1_id,
                "timeseries_id": ts_1_id,
            }
            ret = client.post(TIMESERIES_BY_SITES_URL, json=tbs_3)
            assert ret.status_code == 403

            # GET by id
            ret = client.get(f"{TIMESERIES_BY_SITES_URL}{tbs_1_id}")
            assert ret.status_code == 200
            tbs_1_etag = ret.headers["ETag"]

            # PUT
            tbs_1["timeseries_id"] = ts_3_id
            ret = client.put(
                f"{TIMESERIES_BY_SITES_URL}{tbs_1_id}",
                json=tbs_1,
                headers={"If-Match": tbs_1_etag},
            )
            assert ret.status_code == 403

            # DELETE
            ret = client.delete(
                f"{TIMESERIES_BY_SITES_URL}{tbs_1_id}",
                headers={"If-Match": tbs_1_etag},
            )
            assert ret.status_code == 403

    def test_timeseries_by_sites_as_anonym_api(
        self, app, sites, timeseries, timeseries_by_sites
    ):
        site_1_id = sites[0]
        ts_1_id = timeseries[0]
        ts_2_id = timeseries[1]
        tbs_1_id = timeseries_by_sites[0]

        client = app.test_client()

        # GET list
        ret = client.get(TIMESERIES_BY_SITES_URL)
        assert ret.status_code == 401

        # POST
        tbs_3 = {
            "site_id": site_1_id,
            "timeseries_id": ts_2_id,
        }
        ret = client.post(TIMESERIES_BY_SITES_URL, json=tbs_3)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(f"{TIMESERIES_BY_SITES_URL}{tbs_1_id}")
        assert ret.status_code == 401

        # PUT
        tbs_1 = {
            "site_id": site_1_id,
            "timeseries_id": ts_1_id,
        }
        ret = client.put(
            f"{TIMESERIES_BY_SITES_URL}{tbs_1_id}",
            json=tbs_1,
            headers={"If-Match": "Dummy-ETag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401

        # DELETE
        ret = client.delete(
            f"{TIMESERIES_BY_SITES_URL}{tbs_1_id}",
            headers={"If-Match": "Dummy-Etag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401



"""Users resources"""
from flask.views import MethodView
from flask_smorest import abort

from bemserver_core.model import User

from bemserver_api import Blueprint
from bemserver_api.database import db

from .schemas import UserSchema, UserQueryArgsSchema, BooleanValueSchema


blp = Blueprint(
    "User", __name__, url_prefix="/users", description="Operations on users"
)


@blp.route("/")
class UserViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.arguments(UserQueryArgsSchema, location="query")
    @blp.response(200, UserSchema(many=True))
    def get(self, args):
        """List users"""
        return User.get(**args)

    @blp.login_required
    @blp.etag
    @blp.arguments(UserSchema)
    @blp.response(201, UserSchema)
    @blp.catch_integrity_error
    def post(self, new_item):
        """Add a new user"""
        password = new_item.pop("password")
        item = User.new(**new_item)
        item.set_password(password)
        item.is_admin = False
        item.is_active = True
        db.session.commit()
        return item


@blp.route("/<int:item_id>")
class UserByIdViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.response(200, UserSchema)
    def get(self, item_id):
        """Get user by ID"""
        item = User.get_by_id(item_id)
        if item is None:
            abort(404)
        return item

    @blp.login_required
    @blp.etag
    @blp.arguments(UserSchema)
    @blp.response(200, UserSchema)
    @blp.catch_integrity_error
    def put(self, new_item, item_id):
        """Update an existing user"""
        item = User.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, UserSchema)
        password = new_item.pop("password")
        item.update(**new_item)
        item.set_password(password)
        db.session.commit()
        return item

    @blp.login_required
    @blp.etag
    @blp.response(204)
    def delete(self, item_id):
        """Delete a user"""
        item = User.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, UserSchema)
        item.delete()
        db.session.commit()


@blp.route("/<int:item_id>/set_admin", methods=("PUT",))
@blp.login_required
@blp.etag
@blp.arguments(BooleanValueSchema)
@blp.response(204)
def set_admin(args, item_id):
    """Set / unset admin"""
    item = User.get_by_id(item_id)
    if item is None:
        abort(404)
    blp.check_etag(item, UserSchema)
    item.is_admin = args["value"]
    db.session.commit()
    blp.set_etag(item, UserSchema)


@blp.route("/<int:item_id>/set_active", methods=("PUT",))
@blp.login_required
@blp.etag
@blp.arguments(BooleanValueSchema)
@blp.response(204)
def set_active(args, item_id):
    """Set / unset active"""
    item = User.get_by_id(item_id)
    if item is None:
        abort(404)
    blp.check_etag(item, UserSchema)
    item.is_active = args["value"]
    db.session.commit()
    blp.set_etag(item, UserSchema)



"""Resources initialization"""

from . import about
from . import users
from . import user_groups
from . import users_by_user_groups
from . import campaigns
from . import user_groups_by_campaigns
from . import campaign_scopes
from . import user_groups_by_campaign_scopes
from . import timeseries_properties
from . import timeseries_data_states
from . import timeseries
from . import timeseries_property_data
from . import timeseries_data
from . import event_levels
from . import event_categories
from . import events
from . import timeseries_by_events
from . import events_by_sites
from . import events_by_buildings
from . import events_by_storeys
from . import events_by_spaces
from . import events_by_zones
from . import sites
from . import buildings
from . import storeys
from . import spaces
from . import zones
from . import structural_element_properties
from . import site_properties
from . import building_properties
from . import storey_properties
from . import space_properties
from . import zone_properties
from . import site_property_data
from . import building_property_data
from . import storey_property_data
from . import space_property_data
from . import zone_property_data
from . import timeseries_by_sites
from . import timeseries_by_buildings
from . import timeseries_by_storeys
from . import timeseries_by_spaces
from . import timeseries_by_zones
from . import energy_sources
from . import energy_end_uses
from . import energy_consumption_timeseries_by_sites
from . import energy_consumption_timeseries_by_buildings
from . import input_output
from . import analysis
from . import st_cleanups_by_campaigns
from . import st_cleanups_by_timeseries
from . import st_check_missings_by_campaigns


MODULES = (
    about,
    users,
    user_groups,
    users_by_user_groups,
    campaigns,
    user_groups_by_campaigns,
    campaign_scopes,
    user_groups_by_campaign_scopes,
    timeseries_properties,
    timeseries_data_states,
    timeseries,
    timeseries_property_data,
    timeseries_data,
    event_levels,
    event_categories,
    events,
    timeseries_by_events,
    events_by_sites,
    events_by_buildings,
    events_by_storeys,
    events_by_spaces,
    events_by_zones,
    sites,
    buildings,
    storeys,
    spaces,
    zones,
    structural_element_properties,
    site_properties,
    building_properties,
    storey_properties,
    space_properties,
    zone_properties,
    site_property_data,
    building_property_data,
    storey_property_data,
    space_property_data,
    zone_property_data,
    timeseries_by_sites,
    timeseries_by_buildings,
    timeseries_by_storeys,
    timeseries_by_spaces,
    timeseries_by_zones,
    energy_sources,
    energy_end_uses,
    energy_consumption_timeseries_by_sites,
    energy_consumption_timeseries_by_buildings,
    input_output,
    analysis,
    st_cleanups_by_campaigns,
    st_cleanups_by_timeseries,
    st_check_missings_by_campaigns,
)


def register_blueprints(api):
    """Initialize application with all modules"""
    for module in MODULES:
        module.register_blueprints(api)



"""Authentication"""
from functools import wraps

import sqlalchemy as sqla
from flask_httpauth import HTTPBasicAuth
from flask_smorest import abort
from bemserver_core.model.users import User
from bemserver_core.authorization import CurrentUser, BEMServerAuthorizationError

from bemserver_api.database import db


class Auth(HTTPBasicAuth):
    """Authentication and authorization management"""

    def login_required(self, f=None, **kwargs):
        """Decorator providing authentication and authorization

        Uses HTTPBasicAuth.login_required authenticate user
        Sets CurrentUser context variable to authenticated user for the request
        Catches Authorization error and aborts accordingly
        """

        def decorator(func):
            @wraps(func)
            def wrapper(*args, **func_kwargs):
                with CurrentUser(self.current_user()):
                    try:
                        resp = func(*args, **func_kwargs)
                    except BEMServerAuthorizationError:
                        abort(403, message="Authorization error")
                return resp

            # Wrap this inside HTTPAuth.login_required
            # to get authenticated user
            return super(Auth, self).login_required(**kwargs)(wrapper)

        if f:
            return decorator(f)
        return decorator


auth = Auth()


@auth.verify_password
def verify_password(username, password):
    """Check password and return User instance"""
    user = db.session.execute(sqla.select(User).where(User.email == username)).scalar()
    if user is not None and user.check_password(password):
        return user
    return None


@auth.error_handler
def auth_error(status):
    """Authentication error handler"""
    # Call abort to trigger error handler and get consistent JSON output
    abort(status, message="Authentication error")



"""Timeseries data tests"""
import contextlib

import pytest

from tests.common import AuthHeader

TIMESERIES_DATA_URL = "/timeseries_data/"
DUMMY_ID = "69"


class TestTimeseriesDataApi:
    @pytest.mark.parametrize("user", ("admin", "user", "anonym"))
    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaigns")
    @pytest.mark.usefixtures("user_groups_by_campaign_scopes")
    @pytest.mark.parametrize("for_campaign", (True, False))
    @pytest.mark.parametrize("format_", ("json", "csv"))
    def test_timeseries_data_get(
        self,
        app,
        user,
        users,
        campaigns,
        timeseries,
        timeseries_data,
        for_campaign,
        format_,
    ):
        start_time, end_time = timeseries_data
        ts_1_id = timeseries[0]
        ts_2_id = timeseries[1]
        campaign_1_id = campaigns[0]
        campaign_2_id = campaigns[1]
        ds_id = 1

        if user == "admin":
            creds = users["Chuck"]["creds"]
            auth_context = AuthHeader(creds)
        elif user == "user":
            creds = users["Active"]["creds"]
            auth_context = AuthHeader(creds)
        else:
            auth_context = contextlib.nullcontext()

        client = app.test_client()

        with auth_context:

            if not for_campaign:
                query_url = TIMESERIES_DATA_URL
                ts_l = (ts_1_id,)
                ret_line_1 = "Datetime,1"
            else:
                query_url = TIMESERIES_DATA_URL + f"campaign/{campaign_1_id}/"
                ts_l = (f"Timeseries {ts_1_id-1}",)
                ret_line_1 = "Datetime,Timeseries 0"

            ret = client.get(
                query_url,
                query_string={
                    "start_time": start_time.isoformat(),
                    "end_time": end_time.isoformat(),
                    "timeseries": ts_l,
                    "data_state": ds_id,
                },
                headers={"Accept": f"application/{format_}"},
            )
            if user == "anonym":
                assert ret.status_code == 401
            else:
                assert ret.status_code == 200
                if format_ == "csv":
                    ret_csv_lines = ret.data.decode("utf-8").splitlines()
                    assert ret_csv_lines[0] == ret_line_1
                    assert ret_csv_lines[1] == "2020-01-01T00:00:00+0000,0.0"
                else:
                    assert list(ret.json.keys()) == [str(ts_l[0])]
                    assert ret.json[str(ts_l[0])] == {
                        "2020-01-01T00:00:00+00:00": 0.0,
                        "2020-01-01T01:00:00+00:00": 1.0,
                        "2020-01-01T02:00:00+00:00": 2.0,
                        "2020-01-01T03:00:00+00:00": 3.0,
                    }

            ret = client.get(
                query_url,
                query_string={
                    "start_time": start_time.isoformat(),
                    "end_time": end_time.isoformat(),
                    "timeseries": ts_l,
                    "data_state": ds_id,
                    "timezone": "Europe/Paris",
                },
                headers={"Accept": f"application/{format_}"},
            )
            if user == "anonym":
                assert ret.status_code == 401
            else:
                assert ret.status_code == 200
                if format_ == "csv":
                    ret_csv_lines = ret.data.decode("utf-8").splitlines()
                    assert ret_csv_lines[0] == ret_line_1
                    assert ret_csv_lines[1] == "2020-01-01T01:00:00+0100,0.0"
                else:
                    assert list(ret.json.keys()) == [str(ts_l[0])]
                    assert ret.json[str(ts_l[0])] == {
                        "2020-01-01T01:00:00+01:00": 0.0,
                        "2020-01-01T02:00:00+01:00": 1.0,
                        "2020-01-01T03:00:00+01:00": 2.0,
                        "2020-01-01T04:00:00+01:00": 3.0,
                    }

            # User not in Timeseries group
            if not for_campaign:
                query_url = TIMESERIES_DATA_URL
                ts_l = (ts_2_id,)
                ret_line_1 = "Datetime,2"
            else:
                query_url = TIMESERIES_DATA_URL + f"campaign/{campaign_2_id}/"
                ts_l = (f"Timeseries {ts_2_id-1}",)
                ret_line_1 = "Datetime,Timeseries 1"

            ret = client.get(
                query_url,
                query_string={
                    "start_time": start_time.isoformat(),
                    "end_time": end_time.isoformat(),
                    "timeseries": ts_l,
                    "data_state": ds_id,
                },
                headers={"Accept": f"application/{format_}"},
            )
            if user == "anonym":
                assert ret.status_code == 401
            elif user == "user":
                assert ret.status_code == 403
            else:
                assert ret.status_code == 200
                if format_ == "csv":
                    ret_csv_lines = ret.data.decode("utf-8").splitlines()
                    assert ret_csv_lines[0] == ret_line_1
                    assert len(ret_csv_lines) > 1
                else:
                    assert list(ret.json.keys()) == [str(ts_l[0])]

            # Unknown timezone
            ret = client.get(
                query_url,
                query_string={
                    "start_time": start_time.isoformat(),
                    "end_time": end_time.isoformat(),
                    "timeseries": ts_l,
                    "data_state": ds_id,
                    "timezone": "DTC",
                },
                headers={"Accept": f"application/{format_}"},
            )
            if user == "anonym":
                assert ret.status_code == 401
            else:
                assert ret.status_code == 422

            # Unknown format
            ret = client.get(
                query_url,
                query_string={
                    "start_time": start_time.isoformat(),
                    "end_time": end_time.isoformat(),
                    "timeseries": ts_l,
                    "data_state": ds_id,
                },
                headers={"Accept": "Dummy"},
            )
            if user == "anonym":
                assert ret.status_code == 401
            else:
                assert ret.status_code == 406

    @pytest.mark.parametrize("user", ("admin", "user", "anonym"))
    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaigns")
    @pytest.mark.usefixtures("user_groups_by_campaign_scopes")
    @pytest.mark.parametrize("for_campaign", (True, False))
    @pytest.mark.parametrize("format_", ("json", "csv"))
    def test_timeseries_data_get_aggregate(
        self,
        app,
        user,
        users,
        campaigns,
        timeseries,
        timeseries_data,
        for_campaign,
        format_,
    ):
        start_time, end_time = timeseries_data
        ts_1_id = timeseries[0]
        ts_2_id = timeseries[1]
        campaign_1_id = campaigns[0]
        campaign_2_id = campaigns[1]
        ds_id = 1

        if user == "admin":
            creds = users["Chuck"]["creds"]
            auth_context = AuthHeader(creds)
        elif user == "user":
            creds = users["Active"]["creds"]
            auth_context = AuthHeader(creds)
        else:
            auth_context = contextlib.nullcontext()

        client = app.test_client()

        with auth_context:

            if not for_campaign:
                query_url = TIMESERIES_DATA_URL
                ts_l = (ts_1_id,)
                ret_line_1 = "Datetime,1"
            else:
                query_url = TIMESERIES_DATA_URL + f"campaign/{campaign_1_id}/"
                ts_l = (f"Timeseries {ts_1_id-1}",)
                ret_line_1 = "Datetime,Timeseries 0"

            ret = client.get(
                f"{query_url}aggregate",
                query_string={
                    "start_time": start_time.isoformat(),
                    "end_time": end_time.isoformat(),
                    "timeseries": ts_l,
                    "data_state": ds_id,
                    "bucket_width_value": 1,
                    "bucket_width_unit": "day",
                    "timezone": "UTC",
                },
                headers={"Accept": f"application/{format_}"},
            )
            if user == "anonym":
                assert ret.status_code == 401
            else:
                assert ret.status_code == 200
                if format_ == "csv":
                    ret_csv_lines = ret.data.decode("utf-8").splitlines()
                    assert ret_csv_lines[0] == ret_line_1
                    assert len(ret_csv_lines) > 1
                else:
                    assert list(ret.json.keys()) == [str(ts_l[0])]

            # Set a bucket width value != 1 for a fixed size period
            if not for_campaign:
                query_url = TIMESERIES_DATA_URL
                ts_l = (ts_1_id,)
                ret_line_1 = "Datetime,1"
            else:
                query_url = TIMESERIES_DATA_URL + f"campaign/{campaign_1_id}/"
                ts_l = (f"Timeseries {ts_1_id-1}",)
                ret_line_1 = "Datetime,Timeseries 0"

            ret = client.get(
                f"{query_url}aggregate",
                query_string={
                    "start_time": start_time.isoformat(),
                    "end_time": end_time.isoformat(),
                    "timeseries": ts_l,
                    "data_state": ds_id,
                    "bucket_width_value": 4,
                    "bucket_width_unit": "hour",
                    "timezone": "UTC",
                },
                headers={"Accept": f"application/{format_}"},
            )
            if user == "anonym":
                assert ret.status_code == 401
            else:
                assert ret.status_code == 200
                if format_ == "csv":
                    ret_csv_lines = ret.data.decode("utf-8").splitlines()
                    assert ret_csv_lines[0] == ret_line_1
                    assert len(ret_csv_lines) > 1
                else:
                    assert list(ret.json.keys()) == [str(ts_l[0])]

            # User not in Timeseries group
            if not for_campaign:
                query_url = TIMESERIES_DATA_URL
                ts_l = (ts_2_id,)
                ret_line_1 = "Datetime,2"
            else:
                query_url = TIMESERIES_DATA_URL + f"campaign/{campaign_2_id}/"
                ts_l = (f"Timeseries {ts_2_id-1}",)
                ret_line_1 = "Datetime,Timeseries 1"

            ret = client.get(
                f"{query_url}aggregate",
                query_string={
                    "start_time": start_time.isoformat(),
                    "end_time": end_time.isoformat(),
                    "timeseries": ts_l,
                    "data_state": ds_id,
                    "bucket_width_value": 1,
                    "bucket_width_unit": "day",
                },
                headers={"Accept": f"application/{format_}"},
            )
            if user == "anonym":
                assert ret.status_code == 401
            elif user == "user":
                assert ret.status_code == 403
            else:
                assert ret.status_code == 200
                if format_ == "csv":
                    ret_csv_lines = ret.data.decode("utf-8").splitlines()
                    assert ret_csv_lines[0] == ret_line_1
                    assert len(ret_csv_lines) > 1
                else:
                    assert list(ret.json.keys()) == [str(ts_l[0])]

            # Unknown format
            ret = client.get(
                f"{query_url}aggregate",
                query_string={
                    "start_time": start_time.isoformat(),
                    "end_time": end_time.isoformat(),
                    "timeseries": ts_l,
                    "data_state": ds_id,
                    "bucket_width_value": 1,
                    "bucket_width_unit": "day",
                },
                headers={"Accept": "Dummy"},
            )
            if user == "anonym":
                assert ret.status_code == 401
            else:
                assert ret.status_code == 406

    @pytest.mark.parametrize("for_campaign", (True, False))
    def test_timeseries_data_get_aggregate_errors(
        self,
        app,
        users,
        campaigns,
        timeseries,
        timeseries_data,
        for_campaign,
    ):
        start_time, end_time = timeseries_data
        campaign_1_id = campaigns[0]
        ts_1_id = timeseries[0]
        ds_id = 1

        creds = users["Chuck"]["creds"]

        client = app.test_client()

        with AuthHeader(creds):

            # Unknown campaign
            if for_campaign:
                query_url = TIMESERIES_DATA_URL + f"campaign/{DUMMY_ID}/"
                ts_l = (f"Timeseries {ts_1_id-1}",)

                ret = client.get(
                    f"{query_url}aggregate",
                    query_string={
                        "start_time": start_time.isoformat(),
                        "end_time": end_time.isoformat(),
                        "timeseries": ts_l,
                        "data_state": ds_id,
                        "bucket_width_value": 1,
                        "bucket_width_unit": "day",
                    },
                )
                assert ret.status_code == 404

            # Unknown data state
            if not for_campaign:
                query_url = TIMESERIES_DATA_URL
                ts_l = (ts_1_id,)
            else:
                query_url = TIMESERIES_DATA_URL + f"campaign/{campaign_1_id}/"
                ts_l = (f"Timeseries {ts_1_id-1}",)

            ret = client.get(
                f"{query_url}aggregate",
                query_string={
                    "start_time": start_time.isoformat(),
                    "end_time": end_time.isoformat(),
                    "timeseries": ts_l,
                    "data_state": DUMMY_ID,
                    "bucket_width_value": 1,
                    "bucket_width_unit": "day",
                },
            )
            assert ret.status_code == 422
            assert ret.json["errors"]["query"]["data_state"] == "Unknown data state ID"

            # Unknown timeseries
            if not for_campaign:
                query_url = TIMESERIES_DATA_URL
                ts_l = (DUMMY_ID,)
            else:
                query_url = TIMESERIES_DATA_URL + f"campaign/{campaign_1_id}/"
                ts_l = ("Dummy name",)

            ret = client.get(
                f"{query_url}aggregate",
                query_string={
                    "start_time": start_time.isoformat(),
                    "end_time": end_time.isoformat(),
                    "timeseries": ts_l,
                    "data_state": ds_id,
                    "bucket_width_value": 1,
                    "bucket_width_unit": "day",
                },
            )
            assert ret.status_code == 422
            assert ret.json["message"].startswith("Unknown timeseries")

            # Unknown aggregation method
            if not for_campaign:
                query_url = TIMESERIES_DATA_URL
                ts_l = (ts_1_id,)
            else:
                query_url = TIMESERIES_DATA_URL + f"campaign/{campaign_1_id}/"
                ts_l = (f"Timeseries {ts_1_id-1}",)

            ret = client.get(
                f"{query_url}aggregate",
                query_string={
                    "start_time": start_time.isoformat(),
                    "end_time": end_time.isoformat(),
                    "timeseries": ts_l,
                    "data_state": ds_id,
                    "bucket_width_value": 1,
                    "bucket_width_unit": "day",
                    "aggregation": "dummy",
                },
            )
            assert ret.status_code == 422

            # Unknown timezone
            ret = client.get(
                f"{query_url}aggregate",
                query_string={
                    "start_time": start_time.isoformat(),
                    "end_time": end_time.isoformat(),
                    "timeseries": ts_l,
                    "data_state": ds_id,
                    "bucket_width_value": 1,
                    "bucket_width_unit": "day",
                    "timezone": "DTC",
                },
            )
            assert ret.status_code == 422

            # Unknown bucket width unit
            if not for_campaign:
                query_url = TIMESERIES_DATA_URL
                ts_l = (ts_1_id,)
            else:
                query_url = TIMESERIES_DATA_URL + f"campaign/{campaign_1_id}/"
                ts_l = (f"Timeseries {ts_1_id-1}",)

            ret = client.get(
                f"{query_url}aggregate",
                query_string={
                    "start_time": start_time.isoformat(),
                    "end_time": end_time.isoformat(),
                    "timeseries": ts_l,
                    "data_state": ds_id,
                    "bucket_width_value": 1,
                    "bucket_width_unit": "dummy",
                    "aggregation": "avg",
                },
            )
            assert ret.status_code == 422

            # Invalid bucket width valid for non fixed size periods
            if not for_campaign:
                query_url = TIMESERIES_DATA_URL
                ts_l = (ts_1_id,)
            else:
                query_url = TIMESERIES_DATA_URL + f"campaign/{campaign_1_id}/"
                ts_l = (f"Timeseries {ts_1_id-1}",)

            ret = client.get(
                f"{query_url}aggregate",
                query_string={
                    "start_time": start_time.isoformat(),
                    "end_time": end_time.isoformat(),
                    "timeseries": ts_l,
                    "data_state": ds_id,
                    "bucket_width_value": 2,
                    "bucket_width_unit": "week",
                    "aggregation": "avg",
                },
            )
            assert ret.status_code == 422

    @pytest.mark.parametrize("user", ("admin", "user", "anonym"))
    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaigns")
    @pytest.mark.usefixtures("user_groups_by_campaign_scopes")
    @pytest.mark.parametrize("for_campaign", (True, False))
    @pytest.mark.parametrize("format_", ("json", "csv"))
    def test_timeseries_data_post(
        self, app, user, users, campaigns, timeseries, for_campaign, format_
    ):

        ts_1_id = timeseries[0]
        ts_2_id = timeseries[1]
        campaign_1_id = campaigns[0]
        campaign_2_id = campaigns[1]
        ds_id = 1

        if user == "admin":
            creds = users["Chuck"]["creds"]
            auth_context = AuthHeader(creds)
        elif user == "user":
            creds = users["Active"]["creds"]
            auth_context = AuthHeader(creds)
        else:
            auth_context = contextlib.nullcontext()

        client = app.test_client()

        with auth_context:

            if not for_campaign:
                query_url = TIMESERIES_DATA_URL
                header = f"Datetime,{ts_1_id}\n"
                ts_l = (ts_1_id,)
            else:
                query_url = TIMESERIES_DATA_URL + f"campaign/{campaign_1_id}/"
                header = f"Datetime,Timeseries {ts_1_id-1}\n"
                ts_l = (f"Timeseries {ts_1_id-1}",)

            if format_ == "csv":
                kwargs = {
                    "data": header
                    + (
                        "2020-01-01T00:00:00+00:00,0\n"
                        "2020-01-01T01:00:00+00:00,1\n"
                        "2020-01-01T02:00:00+00:00,2\n"
                        "2020-01-01T03:00:00+00:00,3\n"
                    )
                }
            else:
                kwargs = {
                    "json": {
                        str(ts_l[0]): {
                            "2020-01-01T00:00:00+00:00": 0,
                            "2020-01-01T01:00:00+00:00": 1,
                            "2020-01-01T02:00:00+00:00": 2,
                            "2020-01-01T03:00:00+00:00": 3,
                        }
                    }
                }
            ret = client.post(
                query_url,
                query_string={
                    "data_state": ds_id,
                },
                **kwargs,
                headers={"Accept": f"application/{format_}"},
            )
            if user == "anonym":
                assert ret.status_code == 401
            else:
                assert ret.status_code == 201
                ret = client.get(
                    query_url,
                    query_string={
                        "start_time": "2020-01-01T00:00:00+00:00",
                        "end_time": "2020-01-01T03:00:00+00:00",
                        "timeseries": ts_l,
                        "data_state": ds_id,
                    },
                )
                assert ret.json

            # User not in Timeseries group
            if not for_campaign:
                query_url = TIMESERIES_DATA_URL
                header = f"Datetime,{ts_2_id}\n"
                ts_l = (ts_2_id,)
            else:
                query_url = TIMESERIES_DATA_URL + f"campaign/{campaign_2_id}/"
                header = f"Datetime,Timeseries {ts_2_id-1}\n"
                ts_l = (f"Timeseries {ts_2_id-1}",)

            if format_ == "csv":
                kwargs = {
                    "data": header
                    + (
                        "2020-01-01T00:00:00+00:00,0\n"
                        "2020-01-01T01:00:00+00:00,1\n"
                        "2020-01-01T02:00:00+00:00,2\n"
                        "2020-01-01T03:00:00+00:00,3\n"
                    )
                }
            else:
                kwargs = {
                    "json": {
                        str(ts_l[0]): {
                            "2020-01-01T00:00:00+00:00": 0,
                            "2020-01-01T01:00:00+00:00": 1,
                            "2020-01-01T02:00:00+00:00": 2,
                            "2020-01-01T03:00:00+00:00": 3,
                        }
                    }
                }
            ret = client.post(
                query_url,
                query_string={
                    "data_state": ds_id,
                },
                **kwargs,
                headers={"Accept": f"application/{format_}"},
            )
            if user == "anonym":
                assert ret.status_code == 401
            elif user == "user":
                assert ret.status_code == 403
            else:
                assert ret.status_code == 201
                ret = client.get(
                    query_url,
                    query_string={
                        "start_time": "2020-01-01T00:00:00+00:00",
                        "end_time": "2020-01-01T03:00:00+00:00",
                        "timeseries": ts_l,
                        "data_state": ds_id,
                    },
                )
                assert ret.json

    @pytest.mark.parametrize("for_campaign", (True, False))
    @pytest.mark.parametrize("format_", ("json", "csv"))
    def test_timeseries_data_post_errors(
        self, app, users, campaigns, timeseries, for_campaign, format_
    ):
        campaign_1_id = campaigns[0]
        ts_1_id = timeseries[0]
        ds_id = 1

        creds = users["Chuck"]["creds"]

        client = app.test_client()

        with AuthHeader(creds):

            # Unknown campaign
            if for_campaign:
                query_url = TIMESERIES_DATA_URL + f"campaign/{DUMMY_ID}/"
                header = f"Datetime,Timeseries {ts_1_id-1}\n"
                ts_l = (f"Timeseries {ts_1_id-1}",)
                if format_ == "csv":
                    kwargs = {"data": header + "2020-01-01T00:00:00+00:00,0\n"}
                else:
                    kwargs = {"json": {str(ts_l[0]): {"2020-01-01T00:00:00+00:00": 0}}}
                ret = client.post(
                    query_url,
                    query_string={
                        "data_state": ds_id,
                    },
                    **kwargs,
                    headers={"Accept": f"application/{format_}"},
                )
                assert ret.status_code == 404

            # Unknown data state
            if not for_campaign:
                query_url = TIMESERIES_DATA_URL
                header = f"Datetime,{ts_1_id}\n"
                ts_l = (ts_1_id,)
            else:
                query_url = TIMESERIES_DATA_URL + f"campaign/{campaign_1_id}/"
                header = f"Datetime,Timeseries {ts_1_id-1}\n"
                ts_l = (f"Timeseries {ts_1_id-1}",)
            if format_ == "csv":
                kwargs = {"data": header + "2020-01-01T00:00:00+00:00,0\n"}
            else:
                kwargs = {"json": {str(ts_l[0]): {"2020-01-01T00:00:00+00:00": 0}}}
            ret = client.post(
                query_url,
                query_string={
                    "data_state": DUMMY_ID,
                },
                **kwargs,
                headers={"Accept": f"application/{format_}"},
            )
            assert ret.status_code == 422
            assert ret.json["errors"]["query"]["data_state"] == "Unknown data state ID"

            # Unknown timeseries
            if not for_campaign:
                query_url = TIMESERIES_DATA_URL
                dummy_ts = DUMMY_ID
                header = f"Datetime,{dummy_ts}\n"
            else:
                query_url = TIMESERIES_DATA_URL + f"campaign/{campaign_1_id}/"
                dummy_ts = "Dummy name"
                header = f"Datetime,{dummy_ts}\n"

            if format_ == "csv":
                kwargs = {"data": header + "2020-01-01T00:00:00+00:00,0\n"}
            else:
                kwargs = {"json": {dummy_ts: {"2020-01-01T00:00:00+00:00": 0}}}
            ret = client.post(
                query_url,
                query_string={
                    "data_state": ds_id,
                },
                **kwargs,
                headers={"Accept": f"application/{format_}"},
            )
            assert ret.status_code == 422
            assert ret.json["message"].startswith("Unknown timeseries")

            # Invalid content
            if not for_campaign:
                query_url = TIMESERIES_DATA_URL
            else:
                query_url = TIMESERIES_DATA_URL + f"campaign/{campaign_1_id}/"
            if format_ == "csv":
                kwargs = {"data": ""}
            else:
                kwargs = {"json": {"dummy": []}}

            ret = client.post(
                query_url,
                query_string={
                    "data_state": ds_id,
                },
                **kwargs,
                headers={"Accept": f"application/{format_}"},
            )
            assert ret.status_code == 422

            # Invalid TS IDs
            if not for_campaign:
                query_url = TIMESERIES_DATA_URL
                if format_ == "csv":
                    kwargs = {
                        "data": "Datetime,Timeseries 1\n2020-01-01T00:00:00+00:00,0\n"
                    }
                else:
                    kwargs = {
                        "json": {"Timeseries 1": {"2020-01-01T00:00:00+00:00": 0}}
                    }

                ret = client.post(
                    query_url,
                    query_string={
                        "data_state": ds_id,
                    },
                    **kwargs,
                    headers={"Accept": f"application/{format_}"},
                )
                assert ret.status_code == 422

            # Unknown format
            ret = client.post(
                query_url,
                query_string={
                    "data_state": ds_id,
                },
                **kwargs,
                headers={"Accept": "Dummy"},
            )
            assert ret.status_code == 406

            # Invalid payload
            ret = client.post(
                query_url,
                query_string={
                    "data_state": ds_id,
                },
                data=bytes.fromhex("2Ef0"),
                headers={"Accept": f"application/{format_}"},
            )
            assert ret.status_code == 422

            # Wrong format
            if format_ == "csv":
                kwargs = {"data": header + "2020-01-01T00:00:00+00:00,0\n"}
            else:
                kwargs = {"json": {str(ts_l[0]): {"2020-01-01T00:00:00+00:00": 0}}}
            wrong_format_ = {"json": "csv", "csv": "json"}[format_]
            ret = client.post(
                query_url,
                query_string={
                    "data_state": ds_id,
                },
                **kwargs,
                headers={"Accept": f"application/{wrong_format_}"},
            )
            assert ret.status_code == 422

    @pytest.mark.parametrize("user", ("admin", "user", "anonym"))
    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaigns")
    @pytest.mark.usefixtures("user_groups_by_campaign_scopes")
    @pytest.mark.parametrize("for_campaign", (True, False))
    def test_timeseries_data_delete(
        self,
        app,
        user,
        users,
        campaigns,
        timeseries,
        timeseries_data,
        for_campaign,
    ):
        start_time, end_time = timeseries_data
        ts_1_id = timeseries[0]
        ts_2_id = timeseries[1]
        campaign_1_id = campaigns[0]
        campaign_2_id = campaigns[1]
        ds_id = 1

        if user == "admin":
            creds = users["Chuck"]["creds"]
            auth_context = AuthHeader(creds)
        elif user == "user":
            creds = users["Active"]["creds"]
            auth_context = AuthHeader(creds)
        else:
            auth_context = contextlib.nullcontext()

        client = app.test_client()

        with auth_context:

            if not for_campaign:
                query_url = TIMESERIES_DATA_URL
                ts_l = (ts_1_id,)
            else:
                query_url = TIMESERIES_DATA_URL + f"campaign/{campaign_1_id}/"
                ts_l = (f"Timeseries {ts_1_id-1}",)

            if user == "admin":
                # Check there is data before deleting
                ret = client.get(
                    query_url,
                    query_string={
                        "start_time": start_time.isoformat(),
                        "end_time": end_time.isoformat(),
                        "timeseries": ts_l,
                        "data_state": ds_id,
                    },
                )
                assert ret.json

            # Delete
            ret = client.delete(
                query_url,
                query_string={
                    "start_time": start_time.isoformat(),
                    "end_time": end_time.isoformat(),
                    "timeseries": ts_l,
                    "data_state": ds_id,
                },
            )
            if user == "anonym":
                assert ret.status_code == 401
            else:
                assert ret.status_code == 204
                # Check data is deleted
                ret = client.get(
                    query_url,
                    query_string={
                        "start_time": start_time.isoformat(),
                        "end_time": end_time.isoformat(),
                        "timeseries": ts_l,
                        "data_state": ds_id,
                    },
                )
                assert not ret.json

            # User not in Timeseries group
            if not for_campaign:
                query_url = TIMESERIES_DATA_URL
                ts_l = (ts_2_id,)
            else:
                query_url = TIMESERIES_DATA_URL + f"campaign/{campaign_2_id}/"
                ts_l = (f"Timeseries {ts_2_id-1}",)

            if user == "admin":
                # Check there is data before deleting
                ret = client.get(
                    query_url,
                    query_string={
                        "start_time": start_time.isoformat(),
                        "end_time": end_time.isoformat(),
                        "timeseries": ts_l,
                        "data_state": ds_id,
                    },
                )
                assert ret.json

            # Delete
            ret = client.delete(
                query_url,
                query_string={
                    "start_time": start_time.isoformat(),
                    "end_time": end_time.isoformat(),
                    "timeseries": ts_l,
                    "data_state": ds_id,
                },
            )
            if user == "anonym":
                assert ret.status_code == 401
            elif user == "user":
                assert ret.status_code == 403
            else:
                assert ret.status_code == 204
                # Check data is deleted
                ret = client.get(
                    query_url,
                    query_string={
                        "start_time": start_time.isoformat(),
                        "end_time": end_time.isoformat(),
                        "timeseries": ts_l,
                        "data_state": ds_id,
                    },
                )
                assert not ret.json

    @pytest.mark.parametrize("method", ("get", "delete"))
    @pytest.mark.parametrize("for_campaign", (True, False))
    def test_timeseries_data_get_delete_errors(
        self,
        app,
        users,
        campaigns,
        timeseries,
        timeseries_data,
        for_campaign,
        method,
    ):
        start_time, end_time = timeseries_data
        campaign_1_id = campaigns[0]
        ts_1_id = timeseries[0]
        ds_id = 1

        creds = users["Chuck"]["creds"]

        client = app.test_client()
        client_method = getattr(client, method)

        with AuthHeader(creds):

            # Unknown campaign
            if for_campaign:
                query_url = TIMESERIES_DATA_URL + f"campaign/{DUMMY_ID}/"
                ts_l = (f"Timeseries {ts_1_id-1}",)

                ret = client_method(
                    query_url,
                    query_string={
                        "start_time": start_time.isoformat(),
                        "end_time": end_time.isoformat(),
                        "timeseries": ts_l,
                        "data_state": ds_id,
                    },
                )
                assert ret.status_code == 404

            # Unknown data state
            if not for_campaign:
                query_url = TIMESERIES_DATA_URL
                ts_l = (ts_1_id,)
            else:
                query_url = TIMESERIES_DATA_URL + f"campaign/{campaign_1_id}/"
                ts_l = (f"Timeseries {ts_1_id-1}",)

            ret = client_method(
                query_url,
                query_string={
                    "start_time": start_time.isoformat(),
                    "end_time": end_time.isoformat(),
                    "timeseries": ts_l,
                    "data_state": DUMMY_ID,
                },
            )
            assert ret.status_code == 422
            assert ret.json["errors"]["query"]["data_state"] == "Unknown data state ID"

            # Unknown timeseries
            if not for_campaign:
                query_url = TIMESERIES_DATA_URL
                ts_l = (DUMMY_ID,)
            else:
                query_url = TIMESERIES_DATA_URL + f"campaign/{campaign_1_id}/"
                ts_l = ("Dummy name",)

            ret = client_method(
                query_url,
                query_string={
                    "start_time": start_time.isoformat(),
                    "end_time": end_time.isoformat(),
                    "timeseries": ts_l,
                    "data_state": ds_id,
                },
            )
            assert ret.status_code == 422
            assert ret.json["message"].startswith("Unknown timeseries")



"""Custom marshmallow fields"""
from pathlib import Path
import json

import marshmallow as ma


TIMEZONES_FILE = Path(__file__).parent / "timezones.json"


with open(TIMEZONES_FILE, encoding="utf-8") as tz_f:
    TIMEZONES = json.load(tz_f)


class Timezone(ma.fields.String):
    """A timezone field

    Validates timezone string is in the list of allowed timezones
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        validator = ma.validate.OneOf(TIMEZONES)
        self.validators.insert(0, validator)
